from astropy.units.quantity import Quantity
from astropy import units as u
from threading import Thread
from queue import Queue
import urllib
import os

PS_URL = 'http://ps1images.stsci.edu/cgi-bin/ps1cutouts?pos=' \
         '{}+%09{}&filter=color&filter=g&filter=r&filter=i&filter=z&filter=y' \
         '&filetypes=stack&auxiliary=data&auxiliary=mask&size={}' \
         '&output_size=0&verbose=0&autoscale=99.500000&catlist='


def get_all_image_urls(ra, dec, size):
    """
    Extract the image urls from the PAN-STARRS image web-site

    :param ra: The RA component of the coordinate of the target
    :type ra: float
    :param dec: The DEC component of the coordinate of the target
    :type dec: float
    :param size: The size of the image in degree
    :type size: float
    :return: Dict with where the keys are the name of the bands and the items are the urls to images
    :rtype: dict
    """

    if ra < 0 or ra > 360 or dec < -30 or dec > 90:
        raise ValueError('Coordinates out of range.\n ra={ra}\tdec{dec}')
    bands = ['g', 'r', 'i', 'z', 'y']
    user_agent = ''.join(['Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_4; en-US) ',
                          'AppleWebKit/534.3 (KHTML, like Gecko) Chrome/6.0.472.63 Safari/534.3'])
    headers = {'User-Agent': user_agent}

    if type(size) == Quantity:
        size = size.to(u.arcsec).value / 0.25
    else:
        size /= 0.25
    size = int(size)
    # create the url of the panstarrs image website
    url = PS_URL.format(ra, dec, size)

    # call the website and download the source code
    req = urllib.request.Request(url, None, headers)
    with urllib.request.urlopen(req) as response:
        page = response.read()
    page = str(page)

    # extract the url of the cutout from the source of the image website
    page = page.split('\\n')
    out = {}
    for band in bands:
        img_url = ''
        bband = ''.join(['.', band, '.'])
        for n in page:
            if 'FITS-cutout' in n and bband in n:
                img_url = n
                break
        img_url = img_url.split('Download FITS cutout')[-1]
        img_url = img_url.split('="')[-1]
        img_url = img_url.split('">')[0]
        out.update({band: img_url})
    return out


def _download_band(url, img_path, results):
    """
    Downloads an image from the panstarrs database

    :param url: The url to the panstarrs image
    :type url: str
    :param img_path: The path the storage place
    :type img_path: str
    :param results: Queue to store the if the image is available or not
    :type results: Queue
    :return:
    """
    try:
        urllib.request.urlretrieve(url, img_path)
        results.put(True)
    except urllib.error.URLError:
        results.put(False)


def download_all_bands(ra, dec, size, save_path):
    """
    Downloads all five filter-images

    :param ra: The ra component of the coordinate in deg
    :type ra: float
    :param dec: The dec component of the coordinate in deg
    :type dec: float
    :param size: The size of the image in degrees
    :type size: float
    :param save_path: The path to the save place with two parameters
    :type save_path: str
    :returns: Dictionary with the band names as keys and the paths to the downloaded images as items
    :rtype: dict
    """
    # check if the directory exists
    if not os.path.exists(save_path):
        os.makedirs(save_path)

    # get the image urls
    img_urls = get_all_image_urls(ra, dec, size)
    ths = []

    path = os.path.join(save_path, 'temp_image_{}.fits')

    out = {}
    q = Queue(maxsize=0)
    # go through all available bands
    for b in img_urls.keys():
        img_path = path.format(b)
        # start a new thread to download every bands
        th = Thread(target=_download_band,
                    args=(f'http:{img_urls[b]}',
                          img_path, q))
        th.start()
        ths.append(th)
        out[b] = img_path

    # wait until all bands are finished or until a timeout of 30sec is reached
    for t in ths:
        t.join(timeout=30)

    while not q.empty():
        if not q.get():
            raise ValueError('Image not available.')
    return out



from astroquery.vizier import Vizier
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table, vstack
from astropy import units as u
from astropy import constants as cons
import numpy as np
import urllib
import numbers
import os

from Phosphorpy.data.sub.spectra import Spectra, SpectraList


class LamostSpectraList(SpectraList):

    def __init__(self, spectra=None):
        SpectraList.__init__(self, spectra)

    def as_dataframe(self, as_velocity=False, redo=False):
        """
        Returns the main information of the stored spectra as a pandas DataFrame

        :param as_velocity:
        :param redo:
        :return: The main information as a dataframe
        :rtype: pandas.DataFrame
        """
        out = []
        for s, d_id in zip(self._spectra, self._ids):
            properties = s.estimate_line_properties(as_velocity=as_velocity,
                                                    redo=redo)
            properties['obsID'] = s.obs_id
            properties['ID'] = d_id
            out.append(properties)
        out = vstack(out).to_pandas()
        return out.set_index('ID')


class LamostSpectra(Spectra):
    obs_id = None
    _wavelength = None
    _flux = None
    header = None
    _lines = None
    estimations = None

    def __init__(self, obsid):
        """
        LamostSpectra provides an interface to work with LAMOST spectra.

        :param obsid: The observation ID of the spectra
        :type obsid: int
        """
        Spectra.__init__(self)
        self.obs_id = obsid
        data, header = download_spectra(obsid)
        self._wavelength = data[2]
        self._flux = data[0]
        self.header = header

    @property
    def radial_velocity(self):
        z = float(self.header['z'])
        if z == -9999:
            z = np.nan
        return (1-z)*cons.c.to(u.km/u.s)

    @property
    def seeing(self):
        return self.header['SEEING']

    @staticmethod
    def from_coordinates(ra, dec, ids=None):
        """
        Queries Lamost spectra from the coordinate(s) and returns a LamostSpectra object in the case
        of a single result and a list of spectra in the case of multiple spectra.

        :param ra: The RA coordinate
        :type ra: float, Union
        :param dec: The Dec coordinate
        :type dec: float, Union
        :param ids: List with ids. ids will be ignored, if ra and dec are floats.
        :type ids: Union
        :return: The results or None, if no spectra is found
        :rtype: LamostSpectra, list
        """
        spectra_list = get_spectra_list_ra_dec(ra, dec)
        if spectra_list is None:
            return None
        elif len(spectra_list) == 1:
            return LamostSpectra(spectra_list['ObsID'][0])
        else:
            out = LamostSpectraList()
            for s in spectra_list:
                if ids is None:
                    out.append(LamostSpectra(s['ObsID']))
                else:
                    out.append(LamostSpectra(s['ObsID']), ids[s['_q']-1])
            return out


def _convert_to_array(x):
    """
    Checks if the input is an numpy array. If not convert it to it
    :param x:
    :return: The input data as numpy array
    :rtype: np.ndarray
    """
    if type(x) != np.ndarray:
        if isinstance(x, numbers.Number):
            x = (x, )
        x = np.array(x)
    return x


def get_spectra_list_ra_dec(ra, dec):
    """
    Starts a query of the LAMOST DR4 catalog and searches for spectra.

    :param ra: RA of the target(s)
    :type ra: float, list, np.ndarray
    :param dec: Dec of the target(s)
    :type dec: float, list, np.ndarray
    :return: A table with the results or None, if no data are found.
    :rtype: Table
    """
    ra = _convert_to_array(ra)
    dec = _convert_to_array(dec)
    return get_spectra_list(SkyCoord(ra*u.deg, dec*u.deg))


def get_spectra_list(coordinates):
    """
    Starts a query of the LAMOST DR4 catalog and searches for spectra.

    :param coordinates: Coordinates of the target(s)
    :type coordinates: SkyCoord
    :return: A table with the results or None, if no data are found.
    :rtype: Table
    """
    if type(coordinates) != SkyCoord:
        raise TypeError('Coordinates must be a astropy.coordinates.SkyCoord object.')

    lamost = Vizier(columns=['_q', 'RAJ2000', 'DEJ2000', 'snru', 'snrg', 'snrr', 'snri', 'snrz',
                             'ObsID', 'z', 'SubClass', 'class'])
    lamost.ROW_LIMIT = -1

    try:
        rs = lamost.query_region(coordinates, 1*u.arcsec, catalog='V/153/dr4')[0]

        return rs
    except IndexError:
        return None


def download_spectra(obsid):
    """
    Downloads a spectra from the LAMOST server

    :param obsid: The observation ID of the spectra.
    :type obsid: int
    :return: The data of the spectra and the header of it
    :rtype: np.ndarray, fits.Header
    """
    temp_path = './temp.fits'
    url = 'http://dr4.lamost.org/./spectrum/fits/{}?token='
    urllib.request.urlretrieve(url.format(obsid),
                               temp_path)
    with fits.open(temp_path) as fi:
        data, header = fi[0].data, fi[0].header

    os.remove(temp_path)
    return data, header



import configparser
import pkg_resources
ADS_LINK = 'https://ui.adsabs.harvard.edu/?#abs/{}'


def create_dict(line):
    """
    Creates a dict from a line of the config file.

    :param line: A line from the config file
    :type line: str
    :return: A dict with one element
    :rtype: dict
    """
    line = line.split(' ')
    temp = [line_part for line_part in line if line_part != '']
    items = temp[1:]
    if len(items) == 1:
        items = items[0]
    return {temp[0]: items}


def read_survey_data():
    """
    Read the survey description from the local survey file

    :return: Dict with all local surveys
    """

    resource_package = 'Phosphorpy'
    resource_path = '/'.join(('local', 'survey.conf'))
    shortcut_path = pkg_resources.resource_filename(resource_package, resource_path)

    conf = configparser.ConfigParser()
    conf.read(shortcut_path)

    surveys = {}
    for s in conf.sections():
        c = conf[s]
        temp = {}
        for k in c:
            line = c[k].split(', ')
            if len(line) > 1:
                temp[k] = line
            else:
                temp[k] = c[k]
        surveys[c['name']] = temp
    return surveys


def add_survey(name, vizier_path, release, reference, magnitudes, coordinates=None, xmatch=None):

    resource_package = 'Phosphorpy'
    resource_path = '/'.join(('local', 'survey.conf'))
    shortcut_path = pkg_resources.resource_filename(resource_package, resource_path)

    conf = configparser.ConfigParser()
    conf.read(shortcut_path)

    sections = list(conf.sections())

    sid = 'survey{}'.format(int(sections[-1].split('survey')[-1])+1)
    conf.add_section(sid)

    conf.set(sid, 'name', value=name)
    conf.set(sid, 'release', release)
    conf.set(sid, 'reference', reference)
    conf.set(sid, 'magnitude', ', '.join(magnitudes))
    conf.set(sid, 'vizier', vizier_path)

    if coordinates is not None:
        conf.set(sid, 'coordinate', ', '.join(coordinates))

    if xmatch is not None:
        conf.set(sid, 'xmatch', xmatch)

    # conf.write()


SURVEY_DATA = read_survey_data()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 09:03:54 2019

@author: Jean Patrick Rauer
"""

from Phosphorpy.core.structure import Table


class Flux(Table):
    
    _fits = None
    _survey = None
    _mask = None
    
    def __init__(self, data, survey=None, mask=None):
        Table.__init__(self, data, survey, mask)

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return str(self.data)

    def get_fluxes(self):
        """
        Returns all fluxes without their errors
        :return:
        """
        names = [n for n in self.data.columns if 'e_' not in n]
        return self.data[names]

    def get_errors(self):
        """
        Returns all errors without their fluxes
        :return:
        """
        names = [n for n in self.data.columns if 'e_' in n]
        return self.data[names]

    def get_index(self):
        """
        Returns the indices of the fluxes

        :return: The indices
        """
        return self.data[[]]

    def get_flux(self, index):
        """
        Returns the fluxes of one source without their errors

        :param index: The index of the source
        :type index: int
        :return:
        """
        return self.get_fluxes().iloc[index]

    def get_error(self, index):
        """
        Returns the errors of a source without the flux

        :param index: The index of the source
        :type index: int
        :return:
        """
        return self.get_errors().iloc[index]

    def get_wavelengths(self):
        pass




from astroquery.irsa_dust import IrsaDust
from astropy.coordinates import SkyCoord
from astropy.table import vstack
from astropy import units as u
import numpy as np
import pandas as pd
import warnings
import numbers


def _to_array(x):
    if type(x) != np.ndarray:
        if isinstance(x, numbers.Number):
            x = np.array([x])
        else:
            x = np.array(x)
    return x


def get_extinctions(ra, dec, wavelengths=None, filter_names=None):
    """
    Fall back method if the extinction package is not available.
    For performance reasons, the extinction in the closest filter is used.

    :param ra: The RA component of the coordinates
    :type ra: Union
    :param dec: The Dec component of the coordinates
    :type dec: Union
    :param wavelengths: The wavelengths of the filter in um or as a astropy Quantity with the dimension length
    :type wavelengths: Union, Quantity
    :param filter_names: The name of the filters
    :type filter_names: Union
    :return: The extinctions of the sources in the different bands
    :rtype: pandas.DataFrame
    """
    warnings.warn('Fall back to direct IRSA dust map queries.\nSlow!!!')
    ra = _to_array(ra)
    dec = _to_array(dec)

    coords = SkyCoord(ra*u.deg, dec*u.deg)

    if wavelengths is None:
        out = []
        for c in coords:
            extinctions = IrsaDust.get_extinction_table(c)
            extinctions['ra'] = c.ra.degree
            extinctions['dec'] = c.dec.degree
            out.append(extinctions)
        return vstack(out).to_pandas()
    else:
        if filter_names is None:
            raise ValueError('If wavelengths are given, filter_names must be given, too.')
        if type(wavelengths) == u.Quantity:
            wavelengths = wavelengths.to(u.um).value

        if isinstance(wavelengths, numbers.Number):
            wavelengths = [wavelengths]
            filter_names = [filter_names]
        out = []
        for c in coords:
            extinctions = IrsaDust.get_extinction_table(c)
            r = []
            for w, names in zip(wavelengths, filter_names):
                z = np.argmin(np.abs(extinctions['LamEff']-w))
                r.append(extinctions['A_SandF'][z])
            out.append(r)
        return pd.DataFrame(data=np.array(out), columns=filter_names)



from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS
from astropy import units as u
from astroquery.sdss import SDSS as sdss
from astroquery.gama import GAMA as GAMA_INTERFACE
from astroquery.vizier import Vizier
import numpy as np
import urllib
import os
import warnings

from Phosphorpy.data.sub.spectra import Spectra, SpectraList


SDSS = 'sdss'
LAMOST = 'lamost'
GAMA = 'gama'
DFGRS = '2fDGR'


def _check_coordinates(coord):
    """
    Checks if the coordinates are scalar and if this is true, convert them to non scalar
    :param coord: The coordinates
    :type coord: SkyCoord
    :return: The input coordinates as non scalar
    :rtype: SkyCoord
    """
    if coord.isscalar:
        coord = SkyCoord(
            np.array([coord.ra.degree])*coord.ra.unit,
            np.array([coord.dec.degree])*coord.dec.unit
        )
    return coord


def _check_ids(ids, coord):
    """
    Checks, if the ids are None and if this is true, create a default ID list.
    Checks also, if the length of the ids is the same as the length coordinates.

    :param ids:
        ID's of the spectra. If no ID's are given, the ID's are set to [0, len(coord)-1].
        Default is None.
    :type ids: Union
    :param coord: The coordinates of the objects for which spectra are wanted
    :type coord: SkyCoord
    :return: The list of ID's
    :rtype: ndarray
    """
    if ids is None:
        ids = np.arange(len(coord))
    else:
        if len(ids) != len(coord):
            raise ValueError('If ID\'s are given, they must have the same length as the given coordinates')
        elif type(ids) != np.ndarray:
            ids = np.array(ids)
    return ids


def get_lamost_spectra(coord, ids=None):
    coord = _check_coordinates(coord)

    ids = _check_ids(ids, coord)

    spec_list = SpectraList()

    lamost_download_url = 'http://dr4.lamost.org/./spectrum/fits/{}?token='
    lamost = Vizier(
        columns=['_q', 'RAJ2000', 'DEJ2000', 'ObsID', 'snru', 'snrg', 'snrr', 'snri', 'snrz', 'z', 'SubClass'])
    lamost.ROW_LIMIT = -1
    rs = lamost.query_region(coord, 1 * u.arcsec, catalog='V/153/dr4')

    if len(rs) == 0:
        return spec_list

    rs = rs[0]

    lamost_coord = SkyCoord(rs['RAJ2000'],
                            rs['DEJ2000'])
    sort = lamost_coord.match_to_catalog_sky(coord)[0]
    ids = ids[sort]

    # create a temporary path with a random number at the end to avoid potential overwriting
    temp_path = f'temp_lamost_spec_{np.random.randint(0, 10000)}.fits'

    for obs_id, index in zip(rs['ObsID'], ids):
        urllib.request.urlretrieve(lamost_download_url.format(obs_id), temp_path)
        with fits.open(temp_path) as fi:
            d = fi[0].data
            wave = d[2]
            fl = d[0]

            spec = Spectra(wavelength=wave,
                           flux=fl,
                           wavelength_unit=u.angstrom,
                           survey=LAMOST,
                           meta=fi[0].header)
            spec_list.append(spec, index)

    os.remove(temp_path)

    return spec_list


def get_sdss_spectra(coord, ids=None):
    """
    Downloads SDSS spectra for objects at the given coordinates, if spectra are available.

    :param coord: The coordinates of the objects for which spectra are wanted
    :type coord: SkyCoord
    :param ids:
        ID's of the spectra. If no ID's are given, the ID's are set to [0, len(coord)-1].
        Default is None.
    :type ids: Union
    :return: SpectraList with the found spectra or an empty SpectraList, if no spectra was found.
    :rtype: SpectraList
    """
    coord = _check_coordinates(coord)

    ids = _check_ids(ids, coord)

    spec_list = SpectraList()
    try:
        rs = sdss.query_region(coord, spectro=True)
    except urllib.error.HTTPError:
        warnings.warn('Connection problem to SDSS')
        rs = None

    # if no SDSS spectra was found
    if rs is None:
        return spec_list
    # convert the output coordinates to SkyCoord and get the order of the input coordinate in
    # respect to tine output coordinates
    sdss_coord = SkyCoord(rs['ra']*u.deg,
                          rs['dec']*u.deg)
    sort = sdss_coord.match_to_catalog_sky(coord)[0]
    ids = ids[sort]

    try:
        # download the SDSS spectra
        sp = sdss.get_spectra(matches=rs)
    except urllib.error.HTTPError:
        warnings.warn('Connection problem to SDSS')
        return spec_list

    # read the spectra and convert the wavelength to a linear scale
    # add the resulting Spectra object to the SpectraList
    for spec, index in zip(sp, ids):
        spec = Table(spec[1].data)
        spec['wavelength'] = np.power(10., spec['loglam'])
        # create a new Spectra object for the spectra with the wavelength and the flux.
        # Use angstrom as default wavelength units
        spec = Spectra(wavelength=spec['wavelength'],
                       flux=spec['flux'],
                       wavelength_unit=u.angstrom,
                       survey=SDSS,
                       meta=spec[1].header)
        spec_list.append(spec, index)

    return spec_list


def get_gama_spectra(coord, ids=None):
    spec_list_gama = SpectraList()
    spec_list_2dfgrs = SpectraList()
    coord = _check_coordinates(coord)

    ids = _check_ids(ids, coord)

    radius = 2./3600
    sql = 'SELECT RA, DEC, URL FROM SpecAll WHERE '
    sql += ' OR '.join(
        [
            f'(RA BETWEEN {ra-radius} AND {ra+radius} AND DEC BETWEEN {dec-radius} AND {dec+radius})'
            for ra, dec in zip(coord.ra.degree, coord.dec.degree)
        ]
    )
    try:
        rs = GAMA_INTERFACE.query_sql(sql)
    except ValueError:
        return spec_list_gama, spec_list_2dfgrs

    gama_coord = SkyCoord(rs['RA']*u.deg, rs['DEC']*u.deg)
    sort = gama_coord.match_to_catalog_sky(coord)[0]
    ids = ids[sort]

    # create a temporary path with a random number at the end to avoid potential overwriting
    temp_path = f'temp_gama_spec_{np.random.randint(0, 10000)}.fits'
    for r, index in zip(rs, ids):
        urllib.request.urlretrieve(r['URL'], temp_path)
        with fits.open(temp_path) as fi:
            # if it is only a PrimaryHDU, then it is 2dFGRS spectra
            if len(fi) == 1:
                header = fi[0].header
                data = fi[0].data
                wavelength = header['CDELT1']*(np.arange(header['NAXIS1'])-header['CRPIX1'])+header['CRVAL1']
                spec = Spectra(wavelength=wavelength,
                               flux=data[0], wavelength_unit=u.angstrom,
                               survey=DFGRS,
                               meta=header)
                spec_list_2dfgrs.append(spec, index)
            else:
                wcs = WCS(fi[0].header)
                waves = wcs.all_pix2world(np.arange(len(fi[0].data[0])), 0, 0)[0]
                waves = np.power(10., waves)
                spec = Spectra(wavelength=waves,
                               flux=fi[0].data[0],
                               wavelength_unit=u.angstrom,
                               survey=GAMA,
                               meta=fi[0].header)
                spec_list_gama.append(spec, index)
    # remove the temporary file
    os.remove(temp_path)
    return spec_list_gama, spec_list_2dfgrs


def get_spectra(coord, ids=None, source='SDSS'):
    """
    Search for spectra of sources at the coordinate(s) in the specific survey

    :param coord: The coordinate(s) of the required sources
    :type coord: SkyCoord
    :param ids:
        ID's of the spectra. If no ID's are given, the ID's are set to [0, len(coord)-1].
        Default is None.
    :type ids: Union
    :param source: The source of the spectra
    :type source: str
    :return: The spectra, if any spectra was found.
    """
    source = source.lower()

    coord = _check_coordinates(coord)
    ids = _check_ids(ids, coord)

    if source == SDSS:
        return get_sdss_spectra(coord, ids=ids)
    elif source == LAMOST:
        return get_lamost_spectra(coord, ids=ids)
    elif source == GAMA:
        return get_gama_spectra(coord, ids=ids)
    else:
        raise ValueError(f'{source} is unknown for spectra.')


def get_all_spectra(coordinates, index=None):
    specs = None
    try:
        specs = get_spectra(coordinates, index)
    except:
        pass

    try:
        lamost = get_spectra(coordinates, index, LAMOST)
        if specs is None:
            specs = lamost
        else:
            specs.merge(lamost)
    except:
        pass

    try:
        gama = get_spectra(coordinates, index, GAMA)
        if specs is None:
            specs = gama[0]
            specs.merge(gama[1])
        else:
            specs.merge(gama[0])
            specs.merge(gama[1])
    except:
        pass

    return specs



import pylab as pl
from astropy import units as u
from astropy.coordinates import Angle


def _angle_plot(x, y, sp, marker, color):
    """

    :param x: X values
    :type x: numpy.ndarray
    :param y: y values
    :type y: numpy.ndarray
    :param sp: subplot environment
    :type sp: Tuple[Any, Any]
    :param marker: Marker type
    :type marker: str
    :param color: color
    :type color: str, None
    :return:
    """
    x = Angle(x*u.deg)
    x = x.wrap_at(180*u.deg)
    y = Angle(y*u.deg)

    sp.scatter(x.radian, y.radian, marker=marker, c=color)


class CoordinatePlot:
    """
    Basic class to provide coordinate plots
    """

    def __init__(self, coordinate):
        self._coordinate = coordinate

    def equatorial(self, path='', marker='.', color='k',
                   mask_colors=None):
        """
        Plot the position of the entries in equatorial coordinates

        :param path: Path the the storage place. An empty string is for showing only.
        :type path: str
        :param marker: The marker type. Default is '.'
        :type marker: str
        :param color: The color of the markers. Default is black
        :type color: str
        :param mask_colors: A color or a list of colors for the mask data
        :type mask_colors: str, list
        :return:
        """
        self._scatter('ra', 'dec', 'R.A. [deg]', 'Dec [deg]',
                      path, marker, color, mask_colors=mask_colors)

    def galactic(self, path='', marker='.', color='k',
                 mask_colors=None):
        """
        Plot the position of the entries in galactic coordinates

        :param path: Path the the storage place. An empty string is for showing only.
        :type path: str
        :param marker: The marker type. Default is '.'
        :type marker: str
        :param color: The color of the markers. Default is black
        :type color: str
        :param mask_colors: A color or a list of colors for the mask data
        :type mask_colors: str, list
        :return:
        """
        self._scatter('l', 'b', 'l [deg]', 'b [deg]',
                      path, marker, color, mask_colors=mask_colors)

    def _scatter(self, col1, col2, x_label, y_label, path, marker, color,
                 mask_colors=None):
        """
        Makes a scatter plot of two coordinates

        :param col1: The first column name of the coordinates
        :type col1: str
        :param col2: The second column name of the coordinates
        :type col2: str
        :param x_label: The x-label
        :type x_label: str
        :param y_label: The y-label
        :type y_label: str
        :param path: The path to the save place
        :type path: str
        :param marker: The marker type
        :type marker: str
        :param color: The color of the markers
        :type color: str
        :param mask_colors: A color or a list of colors for the mask data
        :type mask_colors: str, list
        :return:
        """
        pl.clf()
        sp = pl.subplot(projection='mollweide')

        _angle_plot(self._coordinate[col1],
                    self._coordinate[col2],
                    sp, marker=marker,
                    color=color)
        if self._coordinate.mask.get_mask_count() > 0:
            # todo: implement color and/or marker iteration
            for mask_id in range(self._coordinate.mask.get_mask_count()):
                _angle_plot(self._coordinate[col1][self._coordinate.mask.get_mask(mask_id)],
                            self._coordinate[col2][self._coordinate.mask.get_mask(mask_id)],
                            sp, marker=marker,
                            color=mask_colors)
        sp.grid(True)
        sp.set_xticklabels(['14h', '16h', '18h', '20h', '22h',
                            '0h', '2h', '4h', '6h', '8h', '10h'])

        sp.set_xlabel(x_label)
        sp.set_ylabel(y_label)

        if path != '':
            pl.savefig(path)
        else:
            pl.show()



"""
This script provides an interface to the CSS-server to download light curves from it.

(http://nesssi.cacr.caltech.edu/DataRelease/)

@author: Jean Patrick Rauer
"""

import requests
import pylab as pl
import urllib
import pandas
import numpy as np
import os
import warnings


def smooth(d, c=5):
    """
    Smooth a 1D data set linearly

    :param d: The input data
    :type d: np.ndarray
    :param c: The number of smooth
    :param c: int
    :return: The c-times smoothed input data
    :rtype: np.ndarray
    """
    if c == 0:
        return d
    x = np.zeros(len(d))
    x[0] = (d[0]+d[1])/2
    x[1:-1] = (2*d[1:-1]+d[2:]+d[:-2])/4
    x[-1] = (d[-2]+d[-1])/2
    return smooth(x, c=c-1)


def smooth_err(d, err, c=5):
    """
    Smooth a 1D data set with respect to the errors

    :param d: The input data
    :type d: np.ndarray
    :param err: The errors of the data
    :type err: np.ndarray
    :param c: The number of smooths
    :type c: int
    :return: The c-times smoothed input data
    :rtype: np.ndarray
    """
    if c == 0:
        return d
    e = 1/err
    x = np.zeros(len(d))
    x[0] = (d[0]*e[0]+d[1]*e[1])/(e[0]+e[1])
    x[1:-1] = (2*d[1:-1]*e[1:-1]+d[2:]*e[2:]+d[:-2]*e[:-2])/(2*e[1:-1]+e[2:]+e[:-2])
    x[-1] = (d[-2]*e[-2]+d[-1]*e[-1])/(e[-2]+e[-1])
    return smooth_err(x, err, c=c-1)


def vari_index(d, err):
    """
    Computes the variability index

    :param d: The input data
    :type d: np.ndarray
    :param err: The errors of the input data
    :type err: np.ndarray
    :return: The variability index
    :rtype: float
    """
    return 1/(len(d)-1)*np.sum(np.square(d-np.mean(d))/err**2)


def download_light_curve(ra, dec):
    """
    Downloads the light curve of the target from the Catalina Sky Survey
    
    :param ra: RA coordinate in degree
    :type ra: float
    :param dec: Dec coordinate in degree
    :type dec: float
    :returns: The light curve of the target
    :rtype: pandas.DataFrame
    """
    try:
        r = requests.post('http://nunuku.caltech.edu/cgi-bin/getcssconedb_release_img.cgi',
                          data={'RA': ra, 'Dec': dec, 'DB': 'photcat', 'OUT': 'csv', 'SHORT': 'short'})
        url = r.text.split('save as to')[-1]
        url = url.split('href=')[-1]
        url = url.split('>download')[0]
        urllib.request.urlretrieve(url, "temp.csv")
        pd = pandas.read_csv('temp.csv')
        os.remove('temp.csv')
        return pd
    except ValueError:
        raise ValueError('No light curve available.')


def download_light_curves(ra, dec):
    """
    Downloads a set of CSS light curves from the CSS server

    :param ra: The RA coordinates
    :type ra: list
    :param dec: The Dec coordinates
    :type dec: list
    :return: The downloaded light curves
    :rtype: pandas.DataFrame
    """
    url = 'http://nesssi.cacr.caltech.edu/cgi-bin/getmulticonedb_release2.cgi'
    url2 = 'http://nesssi.cacr.caltech.edu/cgi-bin/getdatamulti.cgi?ID={}&txtInput=0000'
    results = []
#    try:
    try:
        part_size = 80
        parts = len(ra)//part_size+1
        for i in range(parts):
            d_ra = ra[i*part_size: (i+1)*part_size]
            d_dec = dec[i*part_size: (i+1)*part_size]
            with open('temp.txt', 'w') as f:
                for j, (r, d) in enumerate(zip(d_ra, d_dec)):
                    row_id = i*part_size+j
                    f.write(f"{row_id}\t{round(r, 5)}\t{round(d, 5)}\n")

            with open('temp.txt') as f:
                r = requests.post(url,
                                  data={'DB': 'photcat', 'OUT': 'csv', 
                                        'SHORT': 'short'},
                                  files={'upload_file': f})

            r = r.text.split('location.href=\'')[-1]
            r = r.split('\'')[0]

            if 'Query service results' in r:
                r = url2.format(
                    r.split('name="ID" value="')[-1].split('"')[0]
                )
                r = requests.get(r).text

                r = r.split('href=')[-1].split('>download')[0]

            urllib.request.urlretrieve(r, "temp.csv")
            pd = pandas.read_csv('temp.csv')
            os.remove('temp.csv')
            os.remove('temp.txt')
            results.append(pd)
    except:
        warnings.warn('Connection problem')

#    except ValueError:
#        raise ValueError('No light curve available.')
    return pandas.concat(results)


def daily_average(d):
    """
    Takes the daily average of the light curve to reduce the noise.
    """
    d = d.copy()
    d['MJD_day'] = np.int32(d['MJD'].values)
    d = d.groupby('MJD_day')
    d = d.aggregate(np.mean)
    return d


def plot_light_curve(ra, dec):
    """
    PLots the CSS light curve of the target
    
    :param ra: RA coordinate in degree
    :type ra: float
    :param dec: Dec coordinate in degree
    :type dec: float
    """
    lc = download_light_curve(ra, dec)
    
    pl.clf()
    sp = pl.subplot(211)
    avg = daily_average(lc)
    sp.errorbar(avg['MJD'],
                avg['Mag'],
                avg['Magerr'],
                fmt='.k',
                capsize=2)
    sp.invert_yaxis()
    sp = pl.subplot(212)
    avg_smooth = smooth_err(avg['Mag'].values, avg['Magerr'].values)
    sp.scatter(avg['MJD'],
               avg_smooth,
               marker='.',
               c='k')
    
    sp.set_xlabel('MJD')
    sp.set_ylabel('mag')
    sp.invert_yaxis()
    pl.show()



import sqlite3

import numpy as np
from astropy.io import fits
from astropy.table import Table

from Phosphorpy.core.structure import Mask


class DataTable:

    _mask = None
    _head = None
    _data = None

    _plot = None
    _q = [0.15, 0.25, 0.75, 0.85]

    _category = None

    def __init__(self, data=None, mask=None, category='table'):
        """
        Basic data table class
        """
        if data is not None:
            self._data = data
            if mask is None:
                mask = Mask(len(self._data))
            self._mask = mask
        elif mask is not None:
            self._mask = mask
        self._category = category

    def set_mask(self, mask):
        if type(mask) != Mask:
            raise ValueError('Mask must be a mask object.')
        self._mask = mask
        # if type(self.data) == list:
        #     for d in self.data:
        #         d.set_mask(mask)

    def stats(self):
        """
        Returns basic statistics (mean, median, std (unbiased), min, max) of the magnitudes

        :return: A DataFrame with the resulting statistics
        :rtype: pandas.core.frame.DataFrame
        """
        st = self.apply([np.mean, np.median, np.std, np.min, np.max])
        st = st.append(self.data.quantile(self.q))
        return st

    def apply(self, func):
        """
        Applies a function to the DataFrame. See `pandas apply method
        <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.apply.html>` for details.

        :param func: The to applied function
        :type: function, list
        :return:
            pandas Series or a DataFrame, in the case of multiple function, with the results of the applied function
        :rtype: pandas.core.series.Series, pandas.core.frame.DataFrame
        """
        return self.data.apply(func)

    def apply_on_ndarray(self, func):
        return func(self._data.values)

    def apply_on_dataframe(self, func):
        """
        Same as :meth:`apply`
        :param func:
        :return:
        """
        return self.apply(func)

    def remove_unmasked_data(self):
        """
        Removes all unmasked (mask[i] == False) from the data
        :return:
        """
        if self._data is not None:
            latest_mask = self._mask.get_latest_mask()
            latest_mask = latest_mask.align(self._data, fill_value=False)[0]

            self._data = self._data[latest_mask]

    def select_nan(self, column):
        """
        Select all rows with a NaN value
        :return:
        """
        if type(column) != str:
            for c in column:
                self.select_nan(c)
        d = self._data[column]
        self.mask.add_mask(d == d, f'Mask all NaN values in {column}')

    def min(self):
        """
        Returns the minimal values of all magnitudes (the magnitudes can come from different sources)

        :return: A pandas series with the column names as indices and the minimal values
        :rtype: pandas.core.series.Series
        """
        return self.apply(np.min)

    def max(self):
        """
        Returns the maximal values of all magnitudes (the magnitudes can come from different sources)

        :return: A pandas series with the column names as indices and the minimal values
        :rtype: pandas.core.series.Series
        """
        return self.apply(np.max)

    @property
    def shape(self):
        return len(self._data), len(self._data.columns)

    @property
    def head(self):
        return self._head

    @property
    def data(self):
        return self._data

    @property
    def plot(self):
        return self._plot

    @plot.setter
    def plot(self, value):
        raise AttributeError('Replacing plot is not allowed!')

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, value):
        if np.min(value) < 0 or np.max(value) > 1:
            raise ValueError('Minimal value must be larger or equal 0 and the maximal value must be smaller or equal 1')
        self._q = value

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask = value

    @property
    def columns(self):
        if type(self._data) == list:
            out = []
            for d in self._data:
                out.extend(d.columns)
        else:
            return self._data.columns

    def to_astropy_table(self, category='table', one=True):
        """
        Returns the data as an astropy.table.Table

        :return: the data
        :rtype: astropy.table.Table
        """
        if one or type(self.data) != list:
            d = Table.from_pandas(self.combine())
            d.meta['category'] = category
        else:
            d = [Table.from_pandas(data) for data in self.data]

        return d

    def to_bin_table_hdu(self, category='table'):
        t = self.to_astropy_table(category, one=False)
        if type(t) == list:
            return [fits.BinTableHDU(k) for k in t]

        return fits.BinTableHDU(t)

    def _write(self, path, function):
        paths = []
        if type(self.data) == list:
            ending = path.split('.')[-1]
            path = path.split(f'.{ending}')[0]
            for d in self.data:
                name = d.survey_name
                p = f'{path}_{name}.{ending}'

                d.__getattr__(function)(p)
                paths.append(p)
        else:
            self.data.__getattr__(function)(path)
            paths.append(path)
        return paths

    def _to_sql(self, path, **kwargs):
        """
        Writes the data to sql database
        :param path: Path to the sql-database
        :type path: str
        :param kwargs: Additional arguments for pandas to_sql
        :return:
        """
        con = sqlite3.Connection(path)
        self._data.to_sql(self._category, con, **kwargs)
        return path

    def write(self, path, data_format='parquet', **kwargs):
        """
        Writes the data to a file

        :param path: Path to the save place
        :type path: str
        :param data_format:
            The format of the data-file. Current supported types are 'parquet', 'csv', 'sql', 'latex' and 'fits'.
        :return:
        """
        data_format = data_format.lower()
        if data_format == 'parquet':
            return self._write(path, 'to_parquet')
        elif data_format == 'csv':
            return self._write(path, 'to_csv')
        elif data_format == 'sql':
            return self._to_sql(path, **kwargs)
        elif data_format == 'latex':
            return self._write(path, 'to_latex')
        elif data_format == 'fits':
            return self.to_astropy_table().write(path, overwrite=True)
        else:
            raise ValueError(f'Format {data_format} is not supported.')

    def combine(self):
        """
        Combines all sub tables into one big table.
        If no sub tables exists, the data are returned
        :return:
        """
        if type(self.data) != list:
            return self.data
        out = None
        for d in self.data:
            if out is None:
                out = d.copy()
            else:
                out = out.join(d)
        return out

    def has_data(self):
        if self._data is not None:
            if type(self._data) == list and len(self._data) == 0:
                return False
            return True
        else:
            return False

    def __str__(self):
        return str(self._data)

    def __getitem__(self, item):
        if type(item) == str:
            return self._data[item]
        return self._data.loc[item]

    def __len__(self):
        return len(self._data)



import pylab as pl
import numpy as np


class SpectraPlot:
    _spectra = None

    def __init__(self, spectra):
        """
        SpectraPlot os the plotting environment of the :class:`Phosphorpy.data.sub.spectra.Spectra` class.

        :param spectra: The spectra
        :type spectra: Phosphorpy.data.sub.spectra.Spectra
        """
        self._spectra = spectra

    def spectra(self, path='', min_wavelength=None, max_wavelength=None, normalize=False):
        """
        Basic spectra plot

        :param path:
            Path, where to save the figure or an empty string, if no save is wanted.
            Default is an empty string.
        :type path: str
        :param min_wavelength:
            The minimal wavelength to plot or None if no limit is wanted.
            Default is None.
        :type min_wavelength: None, float
        :param max_wavelength:
            The maximal wavelength to plot or None, if no limit is wanted.
        :type max_wavelength: None, float
        :return:
        """
        pl.clf()
        sp = pl.subplot()

        wave = self._spectra.wavelength
        flux = self._spectra.flux

        if normalize:
            flux /= np.nanmedian(flux)

        if min_wavelength is not None or max_wavelength is not None:
            spec = self._spectra.cut(min_wavelength=min_wavelength,
                                     max_wavelength=max_wavelength,
                                     inplace=False)
            wave = spec.wavelength
            flux = spec.flux

        m = np.abs(flux) < 1e8
        sp.step(wave[m], flux[m], '-k')

        if len(self._spectra.fit) > 0:
            for fit in self._spectra.fit:
                sp.plot(wave, fit(wave), '--')

        sp.set_xlabel(f'wavelength [{self._spectra.wavelength_unit.to_string("latex")}]')
        sp.set_ylabel(f'flux')

        sp.set_xlim(wave.min(), wave.max())

        if path == '':
            pl.show()
        else:
            pl.savefig(path)


class SpectraListPlot:
    _spectra_list = None

    def __init__(self, spectra_list):
        """
        SpectraPlot os the plotting environment of the :class:`Phosphorpy.data.sub.spectra.Spectra` class.

        :param spectra_list: The spectra
        :type spectra_list: Phosphorpy.data.sub.spectra.SpectraList
        """
        self._spectra_list = spectra_list

    def spectra(self, index, path='', min_wavelength=None, max_wavelength=None, normalize=False):
        if type(index) == int:
            self._spectra_list.get_by_id(index)[0][0].plot.spectra(path, min_wavelength, max_wavelength)
        elif type(index) == list:
            pl.clf()
            sp = pl.subplot()
            for i in index:
                specs = self._spectra_list.get_by_id(i)
                for spec_id in range(len(specs)):
                    spec = specs[spec_id]
                    if type(spec) == tuple:
                        spec = spec[0]
                    wave = spec.wavelength
                    flux = spec.flux
                    if min_wavelength is not None:
                        m = wave >= min_wavelength
                        wave = wave[m]
                        flux = flux[m]

                    if max_wavelength is not None:
                        m = wave <= max_wavelength
                        wave = wave[m]
                        flux = flux[m]
                    m = np.abs(flux) < 1e8
                    if normalize:
                        flux /= np.nanmedian(flux[m])
                    sp.step(wave[m], flux[m], label=f'{i}')

            sp.set_xlabel('wavelength')
            sp.set_ylabel('flux')
            pl.legend(loc='best')
            pl.show()



from .survey_data import SURVEY_DATA

# Set the names of the surveys
# TODO: use a file instead of hard coding
PANSTARRS = 'PAN-STARRS'
SDSS = 'SDSS'
WISE = 'WISE'
TWO_MASS = '2MASS'
APASS = 'APASS'
KIDS = 'KiDS'
VIKING = 'VIKING'
GALEX = 'GALEX'
GAIA = 'GAIA'
UKIDSS = 'UKIDSS'
BAILER_JONES = 'Bailer-Jones'
SkyMapper = 'SkyMapper'

# Dict where the keys are the survey names and
# the items are a list of the corresponding filter names
MAG_COLUMNS = {}
for k in SURVEY_DATA:
    MAG_COLUMNS[k] = SURVEY_DATA[k]['magnitude']


def get_available_surveys():
    """
    Returns a list with the available surveys

    :return: The name of the surveys
    :rtype: list
    """
    surveys = list(MAG_COLUMNS.keys())
    return surveys


def get_survey_magnitude_names(survey_name):
    """
    Returns a list with the filter names of the survey

    :param survey_name: The name of the survey
    :type survey_name: str
    :return: List with the filter names
    :rtype: list
    """
    if survey_name in MAG_COLUMNS.keys():
        return MAG_COLUMNS[survey_name]
    else:
        raise AttributeError('No survey with name \'{}\' is in the list!'.format(survey_name))




from threading import Thread

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table, vstack
from astroquery.vizier import Vizier as Viz
from astroquery.simbad import Simbad
from pandas import DataFrame
from ..config.survey_data import SURVEY_DATA
from ..config import names
from .xmatch import xmatch
from ..config.match import match_catalogs

import numpy as np
import warnings


warnings.simplefilter('ignore')

Viz.ROW_LIMIT = -1
Simbad.add_votable_fields('otype')


class Vizier:
    columns = ['*']
    catalog = ''
    radius = 2*u.arcsec
    vizier = Viz
    name = ''
    ra_name = 'RAJ2000'
    dec_name = 'DEJ2000'
    cache = None
    reference = ''
    release = ''

    def __init__(self, catalog='', name=None):
        """
        Basic class for an interface between the Vizier class of the astroquery package

        :param catalog: The name of the catalog in the Vizier-database
        :type catalog: str
        """
        if name is not None:
            self.name = name
        self.catalog = catalog
        selected_survey = SURVEY_DATA[self.name]
        self.reference = selected_survey['reference']
        self.catalog = selected_survey['vizier']
        self.release = selected_survey['release']
        cols = []

        # check if coordinate names are set
        # if yes, take set coordinate names
        if 'coordinate' in selected_survey.keys():
            coordinate_names = selected_survey['coordinate']
        # if not take the default ones
        else:
            coordinate_names = [self.ra_name, self.dec_name]
        cols.extend(coordinate_names)

        # transform the filter names to the vizier style
        # except it is GALEX
        if self.name != names.GALEX:
            for c in selected_survey['magnitude']:
                cols.append(f'{c}mag')
                cols.append(f'e_{c}mag')
        else:
            for c in selected_survey['magnitude']:
                cols.append(c)
                cols.append(f'e_{c}')

        # if there are additional columns set
        if 'columns' in selected_survey.keys():
            cols.extend(selected_survey['columns'])
        self.columns = cols

    def __adjust_coordinate_names__(self):
        """
        Reformat the coordinate labeling to avoid troubles with similar names
        :return:
        """
        if self.name != '' and self.name not in self.ra_name:
            self.ra_name = f'{self.name}_{self.ra_name}'
            self.dec_name = f'{self.name}_{self.dec_name}'

    def __query__(self, coordinates):
        """
        Perform the query with a split of the coordinates in 200 item
        blocks to reduce a problem with the connection

        :param coordinates: The list of coordinates of the targets
        :type coordinates: SkyCoord
        :return: A array or a list of arrays with the results of the queries
        :rtype: list, numpy.ndarray
        """

        self.__adjust_coordinate_names__()
        # set how many coordinates should be queried at the same time
        steps = 200
        out = []

        if type(self.catalog) == list:
            for i in range(len(self.catalog)):
                out.append([])
        else:
            out = [[]]

        if coordinates.isscalar:
            coordinates = SkyCoord([coordinates.ra.degree],
                                   [coordinates.dec.degree],
                                   unit=(u.deg, u.deg))
        length_coordinates = len(coordinates)

        # perform the query with maximal 200 targets at the same time
        # this is necessary because otherwise it happens from time to time that a timeout appears
        i = 0
        while i*steps < length_coordinates:
            coords = coordinates[i*steps: (i+1)*steps]
            rs = self.vizier.query_region(coords, self.radius,
                                          catalog=self.catalog)
            for j, r in enumerate(rs):
                out[j].append(r)
            i += 1

        # stack the results
        for i, out_sec in enumerate(out):
            if len(out_sec) != 0:
                o = vstack(out_sec)

                if self.name == 'GAIA':
                    ra_name = 'RA_ICRS'
                    dec_name = "DE_ICRS"
                elif self.name == 'SDSS' or self.name == 'SkyMapper':
                    ra_name = '_RAJ2000'
                    dec_name = '_DEJ2000'
                else:
                    ra_name = 'RAJ2000'
                    dec_name = 'DEJ2000'

                o.rename_column('_q', 'id')
                o.rename_column(ra_name, 'ra')
                o.rename_column(dec_name, 'dec')
                o = o.to_pandas()
                o = o.groupby('id')
                o = o.aggregate(np.nanmean)

                out[i] = o
        # print(out)
        # if only one catalog query is performed
        if len(out) == 1:
            # out = out[0]
            self.cache = out
            return out
        else:
            self.cache = out
            return out

    def __query_single__(self, coord):
        """
        Perform a query of a single catalog

        :param coord: The coordinates of the target(s)
        :type coord: astropy.coordinates.SkyCoord
        :return: The results of the query
        :rtype: astroquery.TableList
        """
        if type(self.radius) == tuple or type(self.radius) == list:
            rs = self.vizier.query_region(coord, width=self.radius[0], height=self.radius[1],
                                          catalog=self.catalog)
        else:
            rs = self.vizier.query_region(coord, self.radius,
                                          catalog=self.catalog)
        return rs

    def query_coordinate(self, coord):
        """
        Performs a query around coordinates

        :param coord: The central coordinates of the area
        :return:
        """
        # check if the coordinates are a tuple or a list
        if type(coord) == tuple or type(coord) == list:
            # if coord has 2 elements, take the first element as RA and the second as Dec
            # todo: add two additional parts with one for radius and one for box size
            if len(coord) == 2:
                ra = coord[0]
                dec = coord[1]
            # if coord has another length than 2, raise an error
            else:
                raise ValueError('If the coordinate are a tuple or list, ' +
                                 f'it must have 2 elements, not {len(coord)}!')
            coord = SkyCoord(ra*u.deg, dec*u.deg)
        elif type(coord) != SkyCoord:
            raise ValueError('"coord" must be a list or a tuple of two elements or a SkyCoord object, '
                             f'not a {type(coord)}')
        elif not coord.isscalar:
            raise ValueError('Only one coordinate set is allowed.')

        return self.__query_single__(coord)[0]

    def query(self, data, ra_name='ra', dec_name='dec', use_xmatch=False, blank=False):
        """
        Start a query around the input positions in the Vizier database

        :param data: The input data
        :type data: numpy.ndarray, astropy.table.Table
        :param ra_name: The name of the RA column
        :type ra_name: str
        :param dec_name: The name of the Dec column
        :type dec_name: str
        :param use_xmatch: True if xmatch should be used, even for small amount of data, else is False.
        :type use_xmatch: bool
        :param blank: True if the raw output should return, else False. Default is False.
        :type blank: bool
        :return: A list of tables with the query results
        :rtype: astropy.table.Table
        """
        if type(data) == DataFrame:
            data = Table.from_pandas(data)
        elif type(data) != Table:
            raise ValueError('Input must be a astropy Table or a pandas DataFrame.')
        # If there are more than 100 entries, use the xmatch interface to get the data
        # this will be faster
        try:
            if len(data) > 0 or use_xmatch:
                out = xmatch(data, ra_name, dec_name, self.name, blank=blank)

                return out
        except ConnectionError:
            if use_xmatch:
                raise ConnectionError('XMatch service does not respond.')
            print('No connection to XMatch, fall back to Vizier.')

        if type(data) == Table:
            # if no units are add to the coordinate columns
            if data[ra_name].unit is None:
                coordinates = SkyCoord(data[ra_name]*u.deg,
                                       data[dec_name]*u.deg)
            # if units are add to the coordinate columns
            else:
                coordinates = SkyCoord(data[ra_name],
                                       data[dec_name])
        # if the data are for example numpy.ndarray
        else:
            coordinates = SkyCoord(np.array(data[ra_name])*u.deg,
                                   np.array(data[dec_name])*u.deg)

        return self.__query__(coordinates)[0]

    def query_constrain(self, **constrains):
        """
        Do a constrain query (see :meth:`astroquery.vizier.Vizier.query_constraints` for details)

        :param constrains: The required constraints
        :type constrains: dict
        :return: The results of the query
        :rtype: astropy.table.Table
        """
        rs = self.vizier.query_constraints(catalog=self.catalog, **constrains)
        return rs[0]


class Gaia(Vizier):
    columns = ['RA_ICRS', 'DE_ICRS',
               'Plx', 'e_Plx',
               'pmRA', 'e_pmRA',
               'pmDE', 'e_pmDE',
               'Gmag', 'e_Gmag',
               'BPmag', 'e_BPmag',
               'RPmag', 'e_RPmag',
               'epsi', 'sepsi']
    name = names.GAIA
    ra_name = 'GAIA_RA_ICRS'
    dec_name = 'GAIA_DE_ICRS'

    def __init__(self):
        """
        Child class to query GAIA DR2 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class TwoMass(Vizier):

    name = names.TWO_MASS

    def __init__(self):
        """
        Child class to query 2MASS data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class SkyMapper(Vizier):

    name = names.SkyMapper

    def __init__(self):
        """
        Child class to query 2MASS data
        """
        Vizier.__init__(self)
        self.columns = ['_RAJ2000', '_DEJ2000',
                        'uPetro', 'e_uPetro', 'vPetro', 'e_vPetro',
                        'gPetro', 'e_gPetro', 'rPetro', 'e_rPetro',
                        'iPetro', 'e_iPetro', 'zPetro', 'e_zPetro']
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class Wise(Vizier):
    name = names.WISE

    def __init__(self):
        """
        Child class to query AllWise data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class PanStarrs(Vizier):
    name = names.PANSTARRS

    def __init__(self):
        """
        Child class to query PanSTARRS DR1 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class SDSS(Vizier):
    name = names.SDSS

    def __init__(self):
        """
        Child class to query SDSS DR12 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class Apass(Vizier):
    name = names.APASS

    def __init__(self):
        """
        Child class to query APASS DR9 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class Galex(Vizier):
    name = names.GALEX

    def __init__(self):
        """
        Child class to query GALEX DR5 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class Kids(Vizier):

    name = names.KIDS

    def __init__(self):
        """
        Child class to query KiDS DR3 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class Viking(Vizier):

    name = names.VIKING

    def __init__(self):
        """
        Child class to query VIKING DR2 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class Ukidss(Vizier):

    name = names.UKIDSS

    def __init__(self):
        """
        Child class to query VIKING DR2 data
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class BailerJones(Vizier):

    name = names.BAILER_JONES

    def __init__(self):
        """
        Child class to query Bailer-Jones distance estimations from the GAIA DR2
        """
        Vizier.__init__(self)
        self.vizier = Viz(columns=self.columns)
        self.vizier.ROW_LIMIT = -1


class MultiSurvey:
    surveys = []

    def __init__(self, sdss=True, panstarrs=True,
                 apass=True, two_mass=True,
                 all_wise=True, gaia=True,
                 galex=True):
        """
        This class provides a simple way to query multiple surveys at the same time.
        It also merges all the resulting data into one file with outer join (no-matches will get empty cells).

        :param sdss: True if the query should include SDSS DR12, else False
        :type sdss: bool
        :param panstarrs: True if the query should include PanSTARRS DR1, else False
        :type panstarrs: bool
        :param apass: True if the query should include APASS DR9, else False
        :type apass: bool
        :param two_mass: True if the query should include 2MASS, else False
        :type two_mass: bool
        :param all_wise: True if the query should include AllWISE, else False
        :type all_wise: bool
        :param gaia: True if the query should include GAIA DR2, else False
        :type gaia: bool
        :param galex: True if the query should include GALEX DR5, else False
        :type galex: bool
        """
        if sdss:
            self.surveys.append(SDSS())

        if panstarrs:
            self.surveys.append(PanStarrs())

        if apass:
            self.surveys.append(Apass())

        if two_mass:
            self.surveys.append(TwoMass())

        if all_wise:
            self.surveys.append(Wise())

        if gaia:
            self.surveys.append(Gaia())

        if galex:
            self.surveys.append(Galex())

    def query(self, data, ra_name='ra', dec_name='dec'):
        """
        Queries the chosen surveys

        :param data: The input data
        :type data: numpy.ndarray, astropy.table.Table
        :param ra_name: The name of the RA column
        :type ra_name: str
        :param dec_name: The name of the Dec column
        :type dec_name: str
        :return: A array with the results of the query
        :rtype: numpy.ndarray
        """
        th = []
        for s in self.surveys:
            t = Thread(target=s.query,
                       args=(data, ra_name, dec_name))
            t.start()
            th.append(t)

        # join the threads
        for t in th:
            t.join()

        data = data.copy()
        # match all the collected data in one table
        for s in self.surveys:
            if len(s.cache) > 0:
                data = match_catalogs(data, s.cache,
                                      ra_name, dec_name,
                                      s.ra_name, s.dec_name)
        return data


def get_survey(name):
    """
    Returns a survey object of the survey with the given name

    :param name: The name of the survey
    :type name: str
    :return: The survey object to perform queries
    :rtype: Vizier
    """
    name = name.lower()
    if name == 'sdss':
        survey = SDSS()
    elif name == 'panstarrs' or name == 'ps' or name == 'pan-starrs':
        survey = PanStarrs()
    elif name == '2mass':
        survey = TwoMass()
    elif name == 'gaia':
        survey = Gaia()
    elif name == 'wise':
        survey = Wise()
    elif name == 'galex':
        survey = Galex()
    elif name == 'apass':
        survey = Apass()
    elif name == 'ukidss':
        survey = Ukidss()
    elif name == 'kids':
        survey = Kids()
    elif name == 'viking':
        survey = Viking()
    elif name == 'bailer-jones' or name == 'bj':
        survey = BailerJones()
    elif name == 'skymapper':
        survey = SkyMapper()
    else:
        raise AttributeError(f'No survey with name {name} known!')
    return survey


def query_by_name(name, data, ra_name='ra', dec_name='dec'):
    """
    Queries a survey by its name

    :param name: Name of the survey
    :type name: str
    :param data: data with the coordinates, it must have at least two columns with 'ra_name' and 'dec_name'
    :type data: astropy.table.Table, pandas.DataFrame
    :param ra_name: Name of the RA column
    :type ra_name: str
    :param dec_name: Name of the Dec column
    :type dec_name: str
    :return: The data from the survey
    :rtype: numpy.ndarray
    """
    survey = get_survey(name)
    out = survey.query(data, ra_name, dec_name)
    return out


def constrain_query(name, **kwargs):
    """
    Constrain query for large surveys by their names

    .. code-block:: python

        results = constrain_query('gaia', Gmag='>16 & <18')

    :param name: The name of the survey
    :type name: str
    :param kwargs: Constrains
    :type kwargs: dict
    :return: The results of the query
    :rtype: astropy.table.Table
    """
    survey = get_survey(name)
    return survey.query_constrain(**kwargs)


def query_simbad(coordinates):
    """
    Queries the Simbad database for potential entries/classifications

    :param coordinates: The coordinates
    :type coordinates: Phosphorpy.data.sub.coordinates.CoordinateTable
    :return:
    """
    sc = coordinates.to_sky_coord()
    rs = Simbad.query_region(sc, radius=2*u.arcsec)
    if rs is None:
        return rs
    coord = []
    ra_str = '{}h{}m{}s'
    dec_str = ' {}d{}m{}s'
    for ra, dec in zip(rs['RA'], rs['DEC']):
        c = ra_str.format(*(ra.split(' ')))
        c += dec_str.format(*(dec.split(' ')))
        coord.append(c)

    s = SkyCoord(np.array(coord))
    rs['ra'] = s.ra
    rs['dec'] = s.dec

    rs = coordinates.match(rs)

    return rs



"""
This script wraps the astroquery.xmatch.XMatch class to create an interface which is more usable in this program.
"""
from astroquery.xmatch import XMatch
from astropy.table import Table, vstack
from astropy import units as u

from ..config.survey_data import SURVEY_DATA
import numpy as np
import pandas
import os


def __check_row_id__(data, colnames=None):
    """
    Checks if the input data has a column 'row_id'. If not it will add such a column with unique integer id's

    :param data: The input data
    :type data: astropy.table.Table, pandas.DataFrame
    :param colnames: List with the names of the columns of the input data
    :type colnames: list
    :return: The input data with an additional column 'row_id' if there wasn't such a column before.
    :rtype: astropy.table.Table, pandas.DataFrame
    """
    if colnames is None:
        if type(data) == pandas.DataFrame:
            colnames = data.columns
        else:
            colnames = data.colnames
    if 'row_id' not in colnames:
        data['row_id'] = np.arange(len(data))+1
    return data


def __write_temp_file__(data, ra_name, dec_name):
    """
    Writes a temporary csv-file with the input data and adds a column with id's if there is no such column

    :param data: The input data
    :type data: astropy.table.Table, pandas.DataFrame
    :param ra_name: The name of the ra column
    :type ra_name: str
    :param dec_name: The name of the Dec column
    :type dec_name: str
    :return:
    """
    coords = ['{}_input'.format(ra_name), '{}_input'.format(dec_name), 'row_id']

    # if the input data are a pandas.DataFrame
    if type(data) == pandas.DataFrame:
        if 'input' not in ra_name:
            data = data.rename({ra_name: '{}_input'.format(ra_name),
                                dec_name: '{}_input'.format(dec_name)},
                               axis='columns')
        cols = list(data.columns)
        data = __check_row_id__(data, cols)
        data.to_csv('temp.csv')
    # if the input data are an astropy.table.Table
    elif type(data) == Table:
        data = data.copy()
        if 'input' not in ra_name:
            data.rename_column(ra_name, '{}_input'.format(ra_name))
            data.rename_column(dec_name, '{}_input'.format(dec_name))
        data = __check_row_id__(data, data.colnames)
        data[coords].write('temp.csv', format='ascii.csv')
    # if the input data aren't an astropy.table.Table or a pandas.DataFrame
    # raise a TypeError
    else:
        raise TypeError('Unsupported data type!')


def __output_columns__(survey):
    """
    Reads the magnitude names and additional columns from the configs and format them in the vizier style.

    :param survey: The name of the survey
    :type survey: str
    :return: The required column names in the vizier style
    :rtype: list
    """
    cols = []
    # add the magnitude column names to the column list
    # use the vizier style (mag_name+mag)
    for c in SURVEY_DATA[survey]['magnitude']:
        cols.append('{}mag'.format(c))
        cols.append('e_{}mag'.format(c))

    # check if additional columns are set. If yes, add them to the column list
    if 'columns' in SURVEY_DATA[survey].keys():
        cols.extend(SURVEY_DATA[survey]['columns'])
    coord_cols = []
    if 'coordinate' in SURVEY_DATA[survey].keys():
        coord_cols.extend(SURVEY_DATA[survey]['coordinate'])
    else:
        coord_cols.extend(['RAJ2000', 'DEJ2000'])
    cols.extend(coord_cols)
    cols.append('row_id')
    return cols, coord_cols


def find_suffix(cols):
    """
    Finds the suffix in the case of non-default labeling

    :param cols: The column labels
    :type cols: Union
    :return: The suffix of the magnitude labels
    :rtype: str
    """
    for c in cols:
        if 'mag' in c:
            pre = c.split('mag')[0]
            return c.split(pre)[-1]
        elif 'Mag' in c:
            if 'Aper' in c:
                pre = c.split('AperMag')[0]
            else:
                pre = c.split('Mag')[0]
            return c.split(pre)[-1]


def _compute_gaia_mags(rs):
    """
    Computes the Gaia magnitudes from the flux errors and rename
    the columns to the Vizier labeling

    :param rs: The gaia data with the flux errors (default gaia DR2 labeling)
    :type rs: pd.DataFrame
    :return: The input DataFrame with the errors and the new column labels
    :rtype: pd.DataFrame
    """
    conv = {'phot_g_mean_mag': "Gmag",
            'phot_bp_mean_mag': 'BPmag',
            'phot_rp_mean_mag': 'RPmag',
            'parallax': 'Plx',
            'parallax_error': 'e_Plx'}
    rs = rs.rename(conv, axis='columns')
    rs['e_Gmag'] = 2.5*rs['phot_g_mean_flux_error']/(np.log(10) *
                                                     rs['phot_g_mean_flux'])
    rs['e_BPmag'] = 2.5*rs['phot_bp_mean_flux_error']/(np.log(10) *
                                                       rs['phot_bp_mean_flux'])
    rs['e_RPmag'] = 2.5*rs['phot_rp_mean_flux_error']/(np.log(10) *
                                                       rs['phot_rp_mean_flux'])
    return rs


def _xmatch(data, ra_name, dec_name, survey, max_distance):
    # write the temp file
    __write_temp_file__(data, ra_name, dec_name)

    # use XMatch to download the data
    rs = XMatch.query(cat1=open('temp.csv'),
                      cat2='vizier:{}'.format(SURVEY_DATA[survey]['vizier']),
                      colRA1='{}_input'.format(ra_name),
                      colDec1='{}_input'.format(dec_name),
                      max_distance=max_distance)

    # remove the temp file
    os.remove('temp.csv')
    return rs


def xmatch(data, ra_name, dec_name, survey, max_distance=1.*u.arcsec, blank=False, max_rows=20000):
    """
    Interface to the astroquery.xmatch.XMatch module

    This function wraps the xmatch function of the astroquery package to make it easier
    to use and to include short-cuts of the survey. As the data input it requires
    a astropy Table or a pandas DataFrame with at least two columns, which are the coordinates.
    The names must be given with the next two parameters.
    The last needed parameter is the name of the survey, which must be specified in the configuration file.

    In the case of a direct use:

    .. code:: python

        from astropy.table import Table
        from Phosphorpy.external.xmatch import xmatch

        sdss_data = xmatch(Table.read('./my_coordinates.fits'), 'ra', 'dec', 'SDSS', blank=True)

    'blank' tell the function to return the complete match data and not the reduced version for the usage
    in **Phosphorpy**.

    :param data: The input data
    :type data: Table, DataFrame
    :param ra_name: The name of the ra column
    :type ra_name: str
    :param dec_name: The name of the Dec column
    :type dec_name: str
    :param survey: The name of the survey
    :type survey: str
    :param max_distance: Maximal distance to the counterpart in the other catalog
    :type max_distance: ﻿astropy.units.quantity.Quantity
    :param blank: True if all columns should return, else False for survey specific columns.
    :type blank: bool
    :param max_rows:
        The number of rows for one query. If more rows than max_rows are in the data, then multiple queries
        are performed with sub-samples of the size <= max_rows.
        Default is 20000, which mostly works. In the case of a slow internet connection and timeout errors,
        a reduction of max_rows can help.
    :type max_rows: int
    :return: The results of the catalog query
    :rtype: pandas.DataFrame
    """
    # if the amount of data is too large for one single query
    if len(data) > max_rows:
        data = __check_row_id__(data)

        rs = []
        steps = len(data)//max_rows
        for i in range(steps):
            rs.append(
                _xmatch(
                    data[i*max_rows: (i+1)*max_rows],
                    ra_name, dec_name, survey, max_distance
                )
            )
        rs = vstack(rs)
    else:
        rs = _xmatch(data, ra_name, dec_name, survey, max_distance)

    if blank:
        return rs.to_pandas()
    # reduce the number of columns to the required ones
    output_cols, coord_cols = __output_columns__(survey)
    # rs = rs[output_cols]

    rs = rs.to_pandas()

    # handle the labeling of XMatch Gaia
    if survey == 'GAIA':
        rs = _compute_gaia_mags(rs)

    else:
        # check if the config-file contains special hints for the labeling
        if 'xmatch' in SURVEY_DATA[survey].keys():
            replacer = SURVEY_DATA[survey]['xmatch']

            # rename the column names to the standard vizier style
            conv = {}
            for c in rs.columns:

                if 'Err' in c:
                    conv[c] = 'e_{}mag'.format(c.replace('{}Err'.format(replacer), '').upper())
                elif 'err' in c:
                    conv[c] = 'e_{}mag'.format(c.replace('{}err'.format(replacer), '').upper())
                elif replacer in c:
                    conv[c] = '{}mag'.format(c.replace(replacer, '').upper())
            rs = rs.rename(conv, axis='columns')

            # special case of UKIDSS J-band
            conv = {}
            for c in rs.columns:

                if '_1' in c:
                    conv[c] = c.replace('_1', '')
            rs = rs.rename(conv, axis='columns')

    # check if output_cols contains the right coordinate names
    if output_cols[-2] not in rs.columns:
        if 'ra' in rs.columns:
            output_cols[-3] = 'ra'
            output_cols[-2] = 'dec'
        elif 'RAdeg' in rs.columns:
            output_cols[-3] = 'RAdeg'
            output_cols[-2] = 'DEdeg'

    rs = rs[output_cols]
    # group by the row id (unique identifier of the input sources)
    # and take mean value of all values without the NaN values
    rs = rs.groupby('row_id')
    rs = rs.aggregate(np.nanmean)

    if type(data) == Table:
        data = data.to_pandas()
        data = __check_row_id__(data, data.columns)
        data.set_index('row_id', inplace=True)
    # make sure that the input and the output data have the same length and the same order
    rs = data[[]].join(rs, how='left')

    return rs



from astropy.table import Table as AstroTable
import numpy as np
import pandas as pd
import os


class Mask:
    _mask = None
    _desc = None

    def __init__(self, length, ids=None):
        if ids is None:
            ids = np.arange(length)+1
        elif len(ids) != length:
            raise AttributeError('"length" and the length of "ids" must be the same value.')

        self._mask = [pd.Series(length*[True], ids)]
        self._desc = ['initialization']

    def __str__(self):
        string = ''
        for i in range(self.get_mask_count()):
            string += f'Mask No {i}: {self.get_description(i)}\n'
        return string

    def add_mask(self, mask, description='', combine=True):
        """
        Adds a new mask to the storage.

        :param mask:
            The new mask with the size of the complete dataset or with the size of passed rows in the previous mask.
        :type mask: pandas.Series
        :param description: The description of the mask. Default is an empty string.
        :type description: str
        :param combine: True if the previous mask should be used (have True values), too, else False. Default is True.
        :type combine: bool
        :return:
        """
        if type(mask) != pd.Series:
            raise ValueError(f'Mask must be a pandas Series but mask has the type {type(mask)}')

        if len(mask) == 0:
            raise ValueError('Must must have at least one element.')
        if combine:
            if len(self._mask) > 0:
                # Aligns the new incoming mask with the previous mask
                # because if a survey doesn't contain all sources the incoming mask
                # will have different size
                # join left because the original mask is the largest by definition
                # fill_value is True because missing values can not be selected
                mask = self._mask[-1].align(mask, fill_value=True, join='left')[1]
                mask &= self._mask[-1]
        else:
            if len(self._mask[0]) != len(mask):
                raise ValueError('New mask must have the same length as the initial one.')

        self._mask.append(mask)
        self._desc.append(description)

    def get_latest_mask(self):
        """
        Returns the latest mask.

        :return: The latest mask
        :rtype: pd.Series
        """
        return self.get_mask(-1)

    def get_latest_description(self):
        """
        Returns the latest description.

        :return: The description text
        :rtype: int
        """
        return self.get_description(-1)

    def get_mask(self, level):
        """
        Returns the description of the mask at the corresponding level.

        :param level: The level of the mask.
        :type level: int
        :return: The mask of the data
        :rtype: pd.Series
        """
        return self._mask[level]

    def get_description(self, level):
        """
        Returns the description of the mask at the corresponding level.

        :param level: The level of the mask.
        :type level: int
        :return: The description of the mask
        :rtype: str
        """
        return self._desc[level]

    def get_mask_count(self):
        """
        Returns the number of masks.

        :return: The number of masks
        :rtype: int
        """
        return len(self._mask)

    def remove_mask(self, mask_id):
        """
        Removes a mask from the list.

        :param mask_id:
        :return:
        """
        self._mask.remove(self._mask[mask_id])
        self._desc.remove(self._desc[mask_id])

    def reset_mask(self):
        """
        Deletes all masks and sets the mask back to initialization status
        :return:
        """
        self._mask = self._mask[:1]
        self._desc = self._desc[:1]

    @property
    def mask(self):
        return self.get_latest_mask()

    @property
    def description(self):
        return self.get_latest_description()

    def __call__(self, *args, **kwargs):
        # if no mask was set
        if len(self._mask) == 0:
            # return the input data
            return args[0]
        # if at least one mask was set
        else:
            # return the mask data
            return args[0][self.get_latest_mask()]

    def __len__(self):
        return len(self._mask)


class Head:
    _name = None

    def __init__(self, name=None):
        if name is None:
            name = ''
        self._name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and type(value) == str:
            self._name = value
        else:
            raise ValueError('New name must be a string.')


class Table:
    __data = None
    __head = None
    __mask = None

    def __init__(self, d, name, mask):
        if d is None:
            d = pd.DataFrame()
        self.__data = d
        self.__head = Head(name=name)

        if mask is None:
            mask = Mask(len(d))
        self.__mask = mask

    def __getitem__(self, item):
        return self.__data[item]

    def __setitem__(self, key, value):
        self.__data[key] = value

    def __len__(self):
        return len(self.__data)

    # def __getattr__(self, item):
    #     if item in dir(self):
    #         return self.__getattr__(item)
    #     return self.__data.__getattribute__(item)

    # def __setattr__(self, key, value):
    #     print('key', key)
    #     if self.__data is None or '_' in key[0] or 'data' in key:
    #         print('to class')
    #         super().__setattr__(key, value)
    #     else:
    #         print('to data')
    #         self.__data.__setattr__(key, value)

    def set_mask(self, mask):
        """
        Sets a new mask to the table

        :param mask: The new mask

        :return:
        """
        self.__mask = mask

    def select_columns(self, columns):
        """
        Select a subset of columns

        :param columns: The name of the columns
        :type columns: Union
        :return:
        """
        self.__data = self.__data[columns]

    def rename(self, name_map, axis=None):
        """
        Rename of columns

        :param name_map: The rename map
        :type name_map: dict
        :param axis:
        :return:
        """
        self.__data.rename(name_map, axis=axis)

    def merge(self, right, left_index=False, right_index=False):
        self.__data = self.__data.merge(right, left_index=left_index, right_index=right_index)

    @property
    def survey_name(self):
        return self.__head.name

    @property
    def mask(self):
        return self.__mask

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, d):
        self.__data = d

    @property
    def survey(self):
        return self.__head

    @property
    def columns(self):
        return self.__data.columns

    def write(self, path, data_format='parquet', **kwargs):
        path, extension = os.path.splitext(path)
        path = f'{path}_{self.survey_name}.{extension}'
        if data_format == 'parquet':
            self.data.to_parquet(path)
        elif data_format == 'sql':
            self.data.to_sql(self.survey_name, kwargs['connection'])
        elif data_format == 'csv':
            self.data.to_csv(path)
        elif data_format == 'fits':
            t = AstroTable.from_pandas(self.data)
            try:
                self.survey.add_to_meta(t.meta, self.survey_name)
            except AttributeError:
                pass
            t.write(path, overwrite=True)
        else:
            raise ValueError('Unknown format')
        return path



import pandas as pd

from Phosphorpy.data.sub.tables.color import Color as ColorTab
from .plots.color import Color
from .table import DataTable


class Colors(DataTable):

    def __init__(self, data=None, mask=None, survey_colors=None):
        DataTable.__init__(self, mask=mask)
        if data is None:
            data = []
        self._data = data

        self._survey_colors = survey_colors

        self._plot = Color(self)

    def __str__(self):
        s = 'Colors:\n'
        for d in self.data:
            s += str(d) + '\n'
        return s[:-2]

    @property
    def survey_colors(self):
        cols = {}
        for d in self.data:
            cols[d.survey_name] = d.columns
        return cols

    def get_survey_data(self, name):
        for d in self.data:
            if d.survey_name == name:
                return d
        raise ValueError(f'No survey with {name} found.')

    def add_colors(self, data, survey_name):
        """
        Add a new ColorTab to the colors

        :param data: The new color table
        :type data: pandas.DataFrame
        :param survey_name: The name of the survey of the colors
        :type survey_name: str
        :return:
        """
        self.data.append(ColorTab(data, survey_name))

    def __get_mask_data__(self, col, minimum, maximum, previous):
        for d in self.data:
            d.set_limit(col, minimum, maximum)

    def set_limit(self, col, minimum=-99, maximum=99, previous=True, survey=None):
        """
        Sets a constrain to the colors and create a new mask of it

        :param col: The columns of the constrain.
        :type col: str, list, tuple
        :param survey:
            Name of the survey or None. If None all surveys with such a color name are used for the limiting
        :type survey: str, None
        :param minimum: The minimal value
        :type minimum: float
        :param maximum: The maximal value
        :type maximum: float
        :param previous: True if the last mask must be True too, else False to create a complete new mask.
        :type previous: bool
        :return:
        """

        if minimum >= maximum:
            raise ValueError(f'Minimum must be bigger than maximum: {minimum} >= {maximum}')
        for d in self.data:
            if survey is None or survey == d.survey_name:
                d.set_limit(col, minimum=minimum, maximum=maximum, previous=previous)

    def get_columns(self, cols):
        """
        Returns all colors with the given color names

        :param cols: The required colors
        :type cols: list, str
        :return: All colors with the given names in it
        :rtype: pd.DataFrame
        """
        print(cols)
        out = None
        for d in self.data:
            try:
                o = d.get_columns(cols)
                if len(o.columns) > 0:
                    if out is None:
                        out = o
                    else:
                        if type(cols) == pd.core.indexes.base.Index:
                            if len(o.columns) > len(out.columns):
                                out = o
                        else:
                            out = out.join(o)
            except ValueError:
                pass
        return out

    def get_column(self, col):
        """
        Returns the color with the given name

        :param col: The required color name
        :type col: str
        :return:
        """
        for d in self.data:
            if col in d.columns.values:
                return d[col]

    def outlier_detection(self, survey):
        """
        Detects outliers in the color data and mask them

        :param survey: The name of the survey where the outlier detection should be performed
        :type survey: str
        :return:
        """
        survey = survey.lower()
        for d in self.data:
            if survey == d.survey_name:
                d.outlier_detection()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 08:43:54 2019

@author: Jean Patrick Rauer
"""
import numpy as np
import pandas as pd

from Phosphorpy.core.functions import power_2_10, subtract
from Phosphorpy.core.structure import Table
from .color import Color
from .flux import Flux


class Magnitude(Table):
    _mag_cols = []
    _err_cols = []
    _survey = None
    
    def __init__(self, data, names, survey='', mask=None):
        Table.__init__(self, data, survey, mask)
        self._set_cols(names)
        self.__remove_sources_without_data()

    def __remove_sources_without_data(self):
        m = np.array([False]*len(self.data))
        for c in self.data.columns:
            d = self.data[c].values
            m |= d == d
        self.data = self.data[m]

    def __str__(self):
        return f'Magnitude of {self.survey_name} with {len(self)} entries\n'

    def set_survey_data(self, survey_data):
        self._survey = survey_data
        
    def _set_cols(self, cols):
        mag = []
        err = []
        for c in cols:
            if 'e_' in c:
                err.append(c)
            else:
                mag.append(c)
        self._mag_cols = mag
        self._err_cols = err

    def apply_extinction_correction(self, correction):
        """
        Applies the extinction correction to the data

        :param correction: A table with the correction values
        :type correction: astropy.table.Table, pandas.DataFrame
        :return:
        """
        rows = np.arange(len(self))
        for c in self._mag_cols:
            self[c] -= correction[c][rows]

    def get_flux(self):
        """
        Returns the flux of the columns
        :return:
        """
        # transform the magnitude in the flux scale
        flux = Flux(self[self._mag_cols].apply(power_2_10).copy(),
                    mask=self.mask, survey=self.survey_name)

        # multiply the flux zero point of all bands of the different surveys
        for m in self._mag_cols:
            fo = self._survey.flux_zero(self.survey_name.lower(), m)
            flux[m] *= fo

        # compute the errors of the fluxes
        errs = self[self._err_cols]
        errs = pd.DataFrame(data=flux.values*errs.values*np.log(10.),
                            columns=errs.columns)

        # return a merged table out of the fluxes and the flux errors
        flux.merge(errs, left_index=True, right_index=True)
        return flux
        
    def get_colors(self, cols=None):
        """
        Computes all possible colors. The first magnitude will be the bluer and the second one will be the redder one,
        except magnitude columns are given, then this order is not fixed.

        .. math:

            C_{i, j} = mag_i - mag_j \\
            \\lambda(band_i) < \\lambda(band_j)

        :param cols:
        :return:
        """
        if cols is None:
            cols = self._mag_cols
        else:
            try:
                cols[0][0][0]
            except TypeError:
                cols = [cols]

        colors = Color(self.data[[]], self.survey_name, mask=self.mask)
        color_color_cols = []
        for i, c1 in enumerate(cols):
            for j, c2 in enumerate(cols):
                
                # compute only such colors where the first argument is in a bluer band
                # (assumption is that the columns are sorted in such a way)
                if i < j:
                    cc_name = '{} - {}'.format(c1, c2)
                    color_color_cols.append(cc_name)
                    if len(self) < 1e4:
                        colors[cc_name] = self[c1].values - self[c2].values
                    else:
                        colors[cc_name] = subtract(self[c1].values, self[c2].values)

        return colors

    def get_columns(self, cols):
        use_columns = []
        for c in self.columns.values:
            if c in cols or c.split('mag')[0] in cols:
                use_columns.append(c)

        return self[use_columns]
    
    def has_full_photometry(self, previous=True):
        """
        Creates a mask of sources, which have photometry values in every band of the survey.

        :param previous: True if the previous mask should be used True, else False. Default is True.
        :type previous: bool
        :return:
        """
        mask = None

        for c in self._mag_cols:

            k = self[c] > -99
            if mask is None:
                mask = k
            else:
                mask = mask & k
        self.mask.add_mask(mask, 'Mask sources with full photometry in {}'.format(self._survey), combine=previous)

    def set_limit(self, band, minimum=99, maximum=-99, previous=True):
        """
        Sets a magnitude limit to the magnitude columns with the name band.

        .. code-block:: python

            # select all sources with a G magnitude between 16 and 18
            # (notice that minimum and maximum are magnitudes, therefore the minimum value is larger than the maximum)
            ds.magnitudes.set_limit('G', minimum=18, maximum=16)


        :param band: The name of the band/magnitude column
        :type band: str
        :param minimum: The minimal magnitude value. Default is 99, which means that no cut will be done.
        :type minimum: float
        :param maximum: The maximal magnitude value. Default is -99, which means that no cut will be done.
        :type maximum: float
        :param previous: True if the previous mask should be used True, else False. Default is True.
        :type previous: bool
        :return:
        """
        if band not in self._mag_cols:
            band_new = band + 'mag'
            if band_new not in self._mag_cols:
                raise ValueError('Magnitude column {} or {} not found.'.format(band, band_new))
            band = band_new
        cut_string = ''
        if minimum != 99:
            cut_string = 'minimum={}'.format(minimum)
        if maximum != -99:
            max_str = 'maximum={}'.format(maximum)
            if cut_string == '':
                cut_string = max_str
            else:
                cut_string = '{} and {}'.format(cut_string, max_str)

        mags = self[band]
        if maximum != -99 and minimum != 99:
            mask = (mags <= minimum) & (mags >= maximum)
        elif maximum != -99:
            mask = mags >= maximum
        elif minimum != 99:
            mask = mags <= 99
        else:
            raise AttributeError('At least one of maximum or minimum must be different to the default value.')
        self.mask.add_mask(mask, 'Apply a magnitude cut on the band {} with {}'.format(band, cut_string),
                           combine=previous)

    @property
    def mag_names(self):
        return self._mag_cols

    @property
    def err_names(self):
        return self._err_cols

    @property
    def survey(self):
        return self._survey



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 09:07:21 2019

@author: Jean Patrick Rauer
"""
import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN

from Phosphorpy.core.structure import Table
from Phosphorpy.data.sub.plots.color import create_color_name


class Color(Table):
    
    _survey = None
    
    def __init__(self, data, name, mask=None):
        Table.__init__(self, data, name, mask=mask)

    def __str__(self):
        return f'Colors:\n{len(self)}\n'

    def __repr__(self):
        return str(self)
        
    def __get_mask_data__(self, col, minimum, maximum, previous):
        if col not in self.columns.values:
            col = create_color_name(col)
        d = self[col]
        mask = (d < maximum) & (d > minimum)
        self.mask.add_mask(mask,
                           f'Color cut (minimum={minimum}, maximum={maximum}',
                           combine=previous)

    def get_columns(self, cols):
        """
        Returns all columns, which are or include a part of the given names.
        If no columns are found, then a ValueError raises.

        :param cols: The name (parts) of the required columns
        :type cols: Union, str
        :return: DataFrame with the required columns
        :rtype: pd.DataFrame
        """
        if type(cols) == str:
            cols = [cols]
        elif type(cols) != list and type(cols) != tuple and type(cols) != set and type(cols) != np.ndarray:
            raise AttributeError('Input must be a string, list, tuple or a set.')

        use_cols = []
        col_names = self.columns
        for c in cols:
            if type(c) != str:
                raise ValueError('Elements must be strings.')
            if c in col_names:
                use_cols.append(c)
            else:
                for column in col_names:
                    if c in column:
                        use_cols.append(column)

        if len(use_cols) == 0:
            raise ValueError('No columns with the given name (parts) found.')

        use_cols = list(np.unique(use_cols))
        return self[use_cols]
        
    def outlier_detection(self):
        """
        Makes an outlier detection based on a DBSCAN.

        NaN values in the colors are replaced with the mean of the color.
        
        .. warning:
            
            In the case of very small amount of data points it can happen
            that all data points are excluded. 
            Use this outlier detection for larger amount of data (N > 1000).
        """
        db = DBSCAN(eps=2, min_samples=5)
        values = self.values.copy()
        for i in range(len(values[0])):
            v = values[:, i]
            v /= np.nanstd(v)
            k = v != v
            v[k] = np.nanmean(v)
            values[:, i] = v
        db.fit(values)
        mask_values = db.labels_ != -1
        mask_values = pd.Series(mask_values, self.index.values)
        self.mask.add_mask(mask_values, 'DBSCAN outlier detection')

    def set_limit(self, col, minimum=-99, maximum=99, previous=True):
        """
        Sets a constrain to the colors and create a new mask of it

        :param col: The columns of the constrain.
        :type col: str, list, tuple
        :param minimum: The minimal value
        :type minimum: float
        :param maximum: The maximal value
        :type maximum: float
        :param previous: True if the last mask must be True too, else False to create a complete new mask.
        :type previous: bool
        :return:
        """
        if type(col) == str:
            self.__get_mask_data__(col, minimum, maximum, previous)
        elif type(col) == list or type(col) == tuple:
            if len(col) == 2:
                for c in self.columns:
                    if col[0] in c and col[1] in c:
                        self.__get_mask_data__(c, minimum, maximum, previous)
                        break
            elif len(col) == 3:
                for c in self.columns:
                    if col[0] in c and col[1] in c and col[2] in c:
                        self.__get_mask_data__(c, minimum, maximum, previous)
                        break
            else:
                raise ValueError('The list must have 2 or 3 elements. Not more and not less.')
        else:
            raise ValueError('Sorry, I don\'t how I should handle col in the format {}. Please Use a string or '
                             'tuple/list to specify the right color')



import warnings

import numpy as np
from astropy import units as u
from astropy.table import Table, join, vstack
from astropy.units.quantity import Quantity
from sklearn.cluster import DBSCAN
from sklearn.neighbors import NearestNeighbors
import pandas

warnings.simplefilter('ignore')


def to_pandas(d):
    """
    Converts the input data to a pandas DataFrame

    :param d: The input data
    :type d: Table, DataFrame
    :return: The input data
    :rtype: DataFrame
    """
    if type(d) == Table:
        return d.to_pandas()
    elif type(d) == pandas.DataFrame:
        return d
    else:
        raise TypeError(f'Unknown type of d ({type(d)}).\nd must be an astropy table or a pandas DataFrame.')


def fill_missing_labels(d, max_label):
    """
    Fills the labels of unmatched sources with unique numbers higher than the highest current label

    :param d: The input data with at least the column 'label'
    :type d: astropy.table.Table
    :param max_label: The current lowest not used label number
    :type max_label: int
    :return: The input data with the new labels and the new lowest not used label number
    :rtype: astropy.table.Table, int
    """
    if type(d) == Table:
        if 'label' not in d.colnames:
            raise ValueError('Missing column "label"!')
    elif type(d) == pandas.DataFrame:
        if 'label' not in d.columns:
            raise ValueError('Missing column "label"!')
    else:
        raise TypeError(f'Unknown type of d ({type(d)}).\nd must be an astropy table or a pandas DataFrame.')

    mask = d['label'] == -1
    unidentified = len(d[mask])
    d['label'][mask] = np.linspace(max_label, max_label + unidentified,
                                   unidentified, dtype=np.int32)
    return d, d['label'].max() + 1


def convert_input_data(d):
    """
    Checks if the input data are a astropy.table.Table and if not it converts it to a Table.

    :param d: The input data
    :type d: numpy.ndarray, astropy.table.Table
    :return: The input data as an astropy.table.Table
    :rtype: astropy.table.Table
    """
    # TODO: reduce to coordinates and id
    if type(d) is not Table:
        if type(d) is pandas.DataFrame:
            d = Table.from_pandas(d)
        else:
            d = Table(d)
    return d


def next_neighbour_id(d1, d2, ra1, dec1, ra2, dec2):
    """
    Returns the id's of the next neighbour

    :param d1: First dataset
    :type d1: Table, DataFrame
    :param d2: Second dataset
    :type d2: Table, DataFrame
    :param ra1: The name of the RA column in the first dataset
    :type ra1: str
    :param dec1: The name of the Dec column in the first dataset
    :type dec1: str
    :param ra2: The name of the RA column in the second dataset
    :type ra2: str
    :param dec2: The name of the Dec column in the second dataset
    :type dec2: str
    :return: Id's of the sources in the first dataset in the order of the second dataset
    :rtype: np.array
    """
    d1 = to_pandas(d1)
    d2 = to_pandas(d2)
    x = np.zeros((len(d1), 2))
    x[:, 0] = d1[ra1].values
    x[:, 1] = d1[dec1].values

    y = np.zeros((len(d2), 2))
    y[:, 0] = d2[ra2].values
    y[:, 1] = d2[dec2].values

    nn = NearestNeighbors(1).fit(x)

    distance, ids = nn.kneighbors(y)
    return ids.reshape((len(d2), ))


def match_catalogs(d1, d2, ra1, dec1, ra2, dec2, join_type='outer',
                   match_radius=2 * u.arcsec):
    """
    Matches two catalog into one
    :param d1: The first input catalog
    :type d1: numpy.ndarray, astropy.table.Table
    :param d2: The second input catalog
    :type d2: numpy.ndarray, astropy.table.Table
    :param ra1: The name of the RA column of the first input catalog
    :type ra1: str
    :param dec1: The name of the Dec column of the first input catalog
    :type dec1: str
    :param ra2: The name of the RA column of the second input catalog
    :type ra2: str
    :param dec2: The name of the Dec column of the second input catalog
    :type dec2: str
    :param join_type: The join mode (inner or outer), default is outer
    :type join_type: str
    :param match_radius: The matching radius in which two sources are considered as the same
    :type match_radius: int, float, astropy.units.quantity.Quantity
    :return: The matched catalog
    :rtype: Table
    """
    # if the match radius is not a Quantity, set the unit to arcsec
    if type(match_radius) is not Quantity:
        match_radius = match_radius * u.arcsec

    len_d1 = len(d1)
    x = np.zeros((len_d1 + len(d2), 2))
    x[:len_d1, 0] = d1[ra1]
    x[:len_d1, 1] = d1[dec1]
    x[len_d1:, 0] = d2[ra2]
    x[len_d1:, 1] = d2[dec2]

    # apply the DBSCAN with the match radius in degree
    db = DBSCAN(eps=match_radius.to(u.deg).value,
                min_samples=2)
    db.fit(x)

    d1 = convert_input_data(d1)
    d2 = convert_input_data(d2)

    d1['label'] = db.labels_[:len_d1]
    d2['label'] = db.labels_[len_d1:]

    max_label = np.max(db.labels_) + 1

    d1, max_label = fill_missing_labels(d1, max_label)
    d2, max_label = fill_missing_labels(d2, max_label)

    d = join(d1, d2, keys='label', join_type=join_type)

    if ra1 == ra2:
        for c in (ra1, dec1):
            d.rename_column(f'{c}_1', c)
            d[c][d.masked] = d[f'{c}_2'][d.masked]
            del d[f'{c}_2']
    else:
        d[ra1][d.masked] = d[ra2][d.masked]
        d[dec1][d.masked] = d[dec2][d.masked]
        del d[ra2]
        del d[dec2]

    # d = np.array(d)
    return d


def group_by_coordinates(d, ra_name, dec_name):
    """
    Groups the input catalog sources by their coordinates in a circle of 2 arcsec and
    perform an nanmean (mean-function with ignoring nan values).

    :param d: The input catalog
    :type d: astropy.table.Table
    :param ra_name: The RA column name
    :type ra_name: str
    :param dec_name: The Dec column name
    :type dec_name: str
    :return: The mean values of sources with multiple entries or the values itself if there is one entry only
    :rtype: astropy.table.Table
    """
    # exclude targets with NAN values in the coordinates (in the stupid way)
    d = d[(d[ra_name] == d[ra_name]) & (d[dec_name] == d[dec_name])]

    x = np.zeros((len(d), 2))
    x[:, 0] = d[ra_name]
    x[:, 1] = d[dec_name]
    db = DBSCAN(eps=3. / 3600,
                min_samples=2)
    db.fit(x)
    d['coord_label'] = db.labels_

    # make a mask of sources which have one entry only
    mask = d['coord_label'] == -1
    out = [d[mask]]

    # select sources with multiple entries
    d = d[np.invert(mask)]

    if len(d) != 0:
        # group the sources with multiple entries by their coordinate label and compute the nanmean value
        d = d.group_by('coord_label')
        d = d.groups.aggregate(np.nanmean)

        # append the results to the output list and merge both tables
        out.append(d)
        out = vstack(out)
    else:
        out = out[0]

    # delete the unused column
    del out['coord_label']
    return out



from matplotlib.cm import get_cmap
import numpy as np
import pylab as pl

SURVEY = ('CSS', 'PTF', 'ZTF')


def _plot_light_curve(sp, lc, label=None, color=None):
    """
    Plots the light curve of a specific target with different markers for every survey

    :param sp: The plotting environment
    :param lc: The light curve data
    :type lc: DataFrame
    :param label: The name of the target
    :type label: str
    :param color: The color of the light curve in the figure
    :return:
    """
    if color is None:
        color = 'k'

    if label is None:
        label = ''

    unique_surveys = np.unique(lc['survey'])
    markers = ['.', 'd', '+', 'x']
    for s in unique_surveys:
        m = markers[s-1]
        l = lc[lc['survey'] == s]
        sp.errorbar(l['mjd'].values,
                    l['mag'].values,
                    l['magerr'].values,
                    fmt=m, capsize=2, alpha=0.2, color=color)
        sp.scatter(l['mjd'].values, l['mag'].values, marker=m,
                   label=f'{label} {SURVEY[s-1]}', color=color)


class LightCurvePlot:

    _light_curve = None

    def __init__(self, light_curve):
        self._light_curve = light_curve

    def plot_light_curve(self, light_curve_id, min_mjd=None, max_mjd=None, path=None):
        """
        Plots the light curve of the CSS

        :param light_curve_id:
            The ID of the light curve source or a tuple of ID's, if multiple light curves should be plotted
        :type light_curve_id: int, tuple, list
        :param path: Path to the storage place. Default is None, which means that the plot will be shown only.
        :type path: str
        :return:
        """

        if min_mjd is None:
            min_mjd = self._light_curve.light_curves['mjd'].min()

        if max_mjd is None:
            max_mjd = self._light_curve.light_curves['mjd'].max()

        pl.clf()
        sp = pl.subplot()
        if type(light_curve_id) is int:
            lc = self._light_curve.light_curves[self._light_curve.light_curves['row_id'] == light_curve_id]
            lc = lc[(lc['mjd'] >= min_mjd) & (lc['mjd'] <= max_mjd)]
            _plot_light_curve(sp, lc)

        elif type(light_curve_id) is tuple or type(light_curve_id) is list:
            colors = get_cmap('Set1').colors
            for i, lci in enumerate(light_curve_id):
                lc = self._light_curve.light_curves[self._light_curve.light_curves['row_id'] == lci]
                lc = lc[(lc['mjd'] >= min_mjd) & (lc['mjd'] <= max_mjd)]
                _plot_light_curve(sp, lc, str(lci),
                                  colors[i % len(colors)])
            pl.legend(loc='best')

        sp.invert_yaxis()

        sp.set_xlabel('MJD [days]')
        sp.set_ylabel('mag')

        if path is not None:
            pl.savefig(path)
        else:
            pl.show()

    def light_curve(self, light_curve_id, min_mjd=None, max_mjd=None, path=None):
        self.plot_light_curve(light_curve_id, min_mjd=min_mjd, max_mjd=max_mjd, path=path)



from datetime import datetime


class MaskLog:
    _mask = None
    _label = ''
    _date = None

    def __init__(self, mask, label):
        self._mask = mask
        self._label = label
        self._date = datetime.now()

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value):
        raise AttributeError('Attribute setting is not allowed')

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = value


class Mask:
    _mask_log = None
    _mask = None

    def __init__(self):
        """
        The Mask-class provides the basic mask system.
        """
        self._mask_log = []

    @property
    def mask(self):
        if self._mask is None:
            raise ValueError('No mask was set.')
        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask_log.append(MaskLog(value, 'new {}'.format(len(self._mask_log)+1)))
        self._mask = value

    def __getitem__(self, item):
        for log in self._mask_log:
            if item == log.label:
                return log
        raise KeyError('Key not found!')

    def __setitem__(self, key, value):
        self._mask_log.append(MaskLog(value, key))



"""
This script provides an interface to download and create color images from SDSS and Pan-STARRS.
These images can be overplotted with markers of the required source and a proper motion arrow.
"""

import os

import numpy as np
import pylab as pl
from astropy import units as u
from astropy.io import fits
from astropy.nddata.utils import Cutout2D
from astropy.visualization import make_lupton_rgb
from astropy.wcs import WCS
from astroquery.sdss import SDSS
import shutil

from Phosphorpy.external.panstarrs import download_all_bands
from Phosphorpy.core.functions import smooth2d


def write(path, c_image, bands):
    """
    Writes the different images to a fits file

    :param path: Path to the images
    :param c_image: the images
    :param bands: the name of the bands
    :return:
    """
    if '.fit' in path:
        path = path.split('.fit')[0]

    if not os.path.exists(os.path.split(path)[0]):
        os.makedirs(os.path.split(path)[0])

    for img, b in zip(c_image, bands):
        img.writeto(f'{path}_{b}.fits', overwrite=True)


def max_count(rgb, center, box_size=2):
    """
    Finds the maximal count in the box for every layer
    :param rgb: The rgb layers
    :type rgb: ndarray
    :param center: The central coordinates (pixel)
    :type center: tuple, list
    :param box_size: The half box size
    :type box_size: bool
    :return: The maximal counts
    :rtype: list
    """
    center_counts = []
    for i in range(3):
        im = rgb[:, :, i].copy()

        med = 2 * np.nanmedian(np.nanmean(im, axis=0))
        im[im < med] = med
        # im -= med

        # collect the center count median
        center_counts.append(np.nanmedian(im[center[0] - box_size: center[0] + box_size,
                                          center[1] - box_size: center[1] + box_size]))
    return center_counts


class Image:
    last_coordinate = None
    color_image_bands = None
    color_image_radius = None
    _available_bands = None

    def _check_bands_(self, bands):
        """
        Checks if the band is a valid PS1 band

        :param bands: The input bands
        :type bands: Union
        :return:

        :raises ValueError: If one of the bands is not included in the valid PS1 bands
        """
        if bands is None:
            return self.color_image_bands
        else:
            if len(bands) != 3:
                raise ValueError("Bands must include three bands (red, green, blue).")
            for b in bands:
                if b not in self._available_bands:
                    raise ValueError(f'{b} is not a valid band. Choose one of {self._available_bands}.')

            return bands

    def _check_size(self, size):
        """
        Checks the size and returns to size in arcmin.

        :param size: The size of the image
        :type size: float, Quantity
        :return: The size of the image in arcmin
        :raises ValueError: If the unit of the size is not equivalent to degree.
        """
        if size is None:
            return self.color_image_radius
        else:
            if type(size) != u.Quantity:
                return size*u.arcmin
            else:
                if size.unit.is_equivalent(u.deg):
                    size = size.to(u.arcmin)

                    if size < 30*u.arcsec:
                        raise ValueError('Size is too small. Minimal size is 30 arcsec.')

                    if size > 30*u.arcmin:
                        raise ValueError('Size is too big. Maximal size is 30 arcmin.')

                    return size
                else:
                    raise ValueError('The unit of \'size\' must be equivalent to degrees.')

    def get_color_image(self, s, path='', bands=None, size=None):
        """
        Interface for the return of a rgb-color image from a survey

        :param s: The central coordinates of the target
        :type s: astropy.coordinates.SkyCoord
        :param path: The path to the storage place.
        :type path: str
        :param bands: The used bands, if others than the default are required
        :type bands: Union
        :param size: The size of the image, if not the default size should be used.
        :type size: float
        :return: The three different color channels
        :rtype: numpy.ndarray
        """
        raise NotImplementedError()


class SDSSImage(Image):

    def __init__(self):
        Image.__init__(self)
        self.last_coordinate = None
        self.color_image_bands = ['z', 'r', 'u']
        self.color_image_radius = 2 * u.arcmin

    def get_color_image(self, s, path='', bands=None, size=None, smooth=2):
        """
        Download the SDSS images and create an RGB image out of them.

        :param s: Coordinates of the target
        :type s: astropy.coordinates.SkyCoord
        :param path:
            The path to the save place. Default is '' which means that the image will be shown only.
        :type path: str
        :param bands: Individual band combination or None. Default is None which means that the default bands are used.
        :type bands: None, tuple, list
        :param size: The wanted size of the image
        :type size: float, astropy.units.Quantity
        :param smooth: Number of smooths
        :type smooth: int
        :return:
        """
        if bands is None:
            bands = self.color_image_bands
        sd = SDSS.get_images(s, band=bands)
        if sd is None:
            return ''
        # if SDSS took more than one image the area
        if len(sd) > 5:
            run = 0
            r_min = 999999

            n_color_bands = len(self.color_image_bands)

            # check only one image per epoch
            for i, hdu in enumerate(sd[::n_color_bands]):
                wcs = WCS(hdu[0].header)
                x, y = wcs.all_world2pix(s.ra.degree, s.dec.degree, 0)
                r = np.hypot(hdu[0].data.shape[0] // 2 - y, hdu[0].data.shape[1] // 2 - x)

                # if the current distance to the center is smaller than the previous distance to the center
                if r < r_min:
                    run = i
                    r_min = r

            # take the images from the closest epoch
            sd = sd[run * n_color_bands: (run + 1) * n_color_bands]

        imgs = []
        head = None
        wcs = None
        for hdu in sd:
            if head is None:
                head = hdu[0].header
                wcs = WCS(head)
                wcs_o = wcs
            else:
                wcs_o = WCS(hdu[0].header)

            data = smooth2d(hdu[0].data, smooth)

            # make a cut around the target
            if size is not None:
                if type(size) != u.Quantity:
                    size = size * u.arcmin
            else:
                size = self.color_image_radius

            cut = Cutout2D(data, s, size, wcs=wcs_o)
            imgs.append(cut.data)

        if '.fit' in path:
            write(path, imgs, bands)
            return

        pl.clf()
        # create a subplot and use the wcs projection
        sp = pl.subplot(projection=wcs)
        rgb = make_lupton_rgb(imgs[2], imgs[1], imgs[0], Q=13,
                              stretch=0.9, minimum=0)
        sp.imshow(rgb, origin='lower')

        sp.set_xlabel('RA')
        sp.set_ylabel('Dec')

        if path == '':
            pl.show()
        else:
            if not os.path.exists(os.path.split(path)[0]):
                os.makedirs(os.path.split(path)[0])
            pl.savefig(path)


class PanstarrsImage(Image):

    def __init__(self):
        """
        Interface to `Pan-STARRS cutout service <https://ps1images.stsci.edu/cgi-bin/ps1cutouts>`_
        """
        Image.__init__(self)
        self.last_coordinate = None
        self.color_image_bands = ['z', 'r', 'g']
        self.color_image_radius = 2 * u.arcmin
        self._available_bands = ['g', 'r', 'i', 'z', 'y']

    def get_normalized_imaged(self, s, smooth, bands=None, size=None, temp_path='./temp/'):
        """
        Returns the normalized images and the HDU's

        :param s: The central coordinates of the image
        :type s: astropy.coordinates.SkyCoord
        :param smooth: The number of linear smooths
        :type smooth: int
        :param bands:
            Optional: Set of three bands for the color image (g, r, i, z, y). Order must be red, green and blue
        :type bands: None, list, tuple
        :param size: Optional: A customized size of the image in arcmin
        :type size: float
        :param temp_path: Path to the temporary storage place. Default is './temp/.
        :type temp_path: str
        :return: The normalized images and the original HDU's
        """
        if smooth < 0 or smooth > 10:
            raise ValueError('\'smooth\' must be between 0 and 10.')

        if type(temp_path) != str:
            raise ValueError('temp_path must be a string.')

        bands = self._check_bands_(bands)

        size = self._check_size(size)

        # download Pan-STARRS images
        paths = download_all_bands(s.ra.degree, s.dec.degree, size,
                                   temp_path)
        imgs = []
        for c in bands:
            imgs.append(fits.open(paths[c]))

        # create an RGB-array
        rgb = np.zeros((imgs[2][0].shape[0], imgs[2][0].shape[1], 3))
        for i in range(3):
            data = imgs[i][0].data
            # med = np.median(data)/2
            # data -= med
            # data[data < 0] = 0
            rgb[:, :, i] = data
        # rgb[:, :, 2] = imgs[2][0].data-
        # rgb[:, :, 1] = imgs[1][0].data
        # rgb[:, :, 0] = imgs[0][0].data
        # take the center
        center = [rgb.shape[0] // 2, rgb.shape[1] // 2]

        # make a lower cut to exclude the noise
        center_counts = max_count(rgb, center)

        if 0 in center_counts:
            center_counts = max_count(rgb, center, rgb.shape[0]//2)

        # normalize to the center median counts
        for i in range(3):
            img = rgb[:, :, i]
            img /= center_counts[i]
            img[img > 100] = 100
            img -= np.nanmin(img)
            img = np.log10(smooth2d(img, smooth))
            img -= np.nanmedian(img)
            img[img < 0] = 0
            img /= np.nanmax(img)

            rgb[:, :, i] = img

        return rgb, imgs

    def get_color_image(self, s, path='', smooth=2, mark_source=False, proper_motion=None, bands=None,
                        size=None):
        """
        Download the Pan-STARRS images and create an RGB image out of them.

        :param s: Coordinates of the target
        :type s: astropy.coordinates.SkyCoord
        :param path:
            The path to the save place. Default is '' which means that the image will be shown only.
        :type path: str
        :param smooth: Number of smooth
        :type smooth: int
        :param mark_source: True if a circle should be plotted around the source, else False. Default is False.
        :type mark_source: bool
        :param proper_motion:
            The proper motion of the source or None. If the proper motion is given, an arrow will indicate
            the proper motion, if None no proper motion arrows will be drawn. Default is None.
        :type proper_motion: None, list
        :param bands: Optional argument with bands other than the defaults (z, r, g)
        :type bands: list
        :param size: The wanted size of the image
        :type size: float, astropy.units.Quantity
        :return:
        """
        temp_path = f'./temp_{np.random.randint(0, 1000000)}/'
        rgb, imgs = self.get_normalized_imaged(s, smooth, bands=bands, size=size,
                                               temp_path=temp_path)

        if '.fit' in path:
            write(path, imgs, self._check_bands_(bands))
            shutil.rmtree(temp_path)
            return

        pl.clf()
        # make a chart with a WCS projection
        sp = pl.subplot(projection=WCS(imgs[0][0].header))
        sp.imshow(rgb, origin='lower')

        # draw the source marker
        if mark_source:
            sp.scatter(s.ra.degree, s.dec.degree, 20, marker='o', c='r')

        # draw the proper motion arrow
        if proper_motion is not None:
            sp.arrow(s.ra.degree, s.dec.degree, proper_motion[0], proper_motion[1], color='r')

        pl.subplots_adjust(top=0.969, bottom=0.118,
                           left=0.146, right=0.973)

        # set the axis labels
        sp.set_xlabel('RA')
        sp.set_ylabel('Dec')

        if path == '':
            pl.show()

        else:
            if not os.path.exists(os.path.split(path)[0]):
                os.makedirs(os.path.split(path)[0])
            pl.savefig(path)
        for i in imgs:
            i.close()
        shutil.rmtree(temp_path)



import pylab as pl


class FluxPlot:
    _data = None

    def __init__(self, data):
        """

        :param data:
        :type data: Phosphorpy.data.sub.flux.FluxTable
        """
        self._data = data

    def sed(self, index, fit=None, path='', x_log=False, y_log=False, legend=False):
        """
        Plot the SED of the source with the given index.

        :param index: The index of the source
        :type index: int
        :param fit:
            A function to fit, a string with the name of a predefined function or an integer for a polynomial fit.
        :type fit: function, str, int
        :param path:
            Path to the place where the figure should be saved. Default is an empty string, which means that
            the figure will be shown only.
        :type path: str
        :param x_log: True if the x-axis should be in log scale, else False. Default is False.
        :type x_log: bool
        :param y_log: True if the y-axis should be in log-scale, else False. Default is False
        :type y_log: bool
        :param legend: True if the legend should be shown, else False. Default is False.
        :type legend: bool
        :return:
        """
        # todo: clean this method
        pl.clf()
        sp = pl.subplot()

        # iterate over all surveys
        for s in self._data.data:
            sp.errorbar(self._data.survey_head.get_survey_wavelengths(s.survey_name),
                        s.get_flux(index),
                        s.get_error(index),
                        fmt='.', capsize=2,
                        label=s.survey_name)

        # data_loc = data_loc[self._data._survey.all_magnitudes()].values
        # mask = data_loc > 0
        # if fit is not None:
        #     if type(fit) == str:
        #         if fit == 'blackbody':
        #             pass
        #         else:
        #             raise ValueError('Unknown function\'{}\''.format(fit))
        #
        #     # if the fit-parameter is a function
        #     # use the function as fitting function
        #     elif callable(fit):
        #         wavelengths = self._data._survey.get_all_wavelengths()
        #         popt, pcov = curve_fit(fit,
        #                                wavelengths[mask],
        #                                data_loc[mask])
        #         waves = np.linspace(np.min(wavelengths),
        #                             np.max(wavelengths),
        #                             1000)
        #         # plot the fitted curve as a black dashed line
        #         sp.plot(waves, fit(waves, *popt), '--k', legend='fit')
        #     elif type(fit) == int:
        #         wavelengths = self._data._survey.get_all_wavelengths()
        #         fit = np.polyfit(wavelengths[mask], data_loc[mask], fit)
        #         poly = np.poly1d(fit)
        #         waves = np.linspace(np.min(wavelengths),
        #                             np.max(wavelengths),
        #                             1000)
        #         sp.plot(waves, poly(waves), '--k')
        #
        #     else:
        #         raise ValueError('Only functions and strings are allowed as fit-parameter values.')

        sp.set_xlabel('wavelength [$\\AA$]')
        sp.set_ylabel('flux')

        if legend:
            pl.legend(loc='best')

        if x_log:
            sp.set_xscale('log')

        if y_log:
            sp.set_yscale('log')

        # if no path is given, show the figure
        if path == '':
            pl.show()
        # if a path is given, save the figure
        else:
            pl.savefig(path)



from setuptools import setup

setup(
    name='Phosphorpy',
    version='0.6.11',
    python_requires='>=3.6',
    packages=['Phosphorpy', 'Phosphorpy.data',
              'Phosphorpy.data.sub', 'Phosphorpy.data.sub.plots',
              'Phosphorpy.config',
              'Phosphorpy.fitting', 'Phosphorpy.external',
              'Phosphorpy.core', 'Phosphorpy.data.sub.tables', 'Phosphorpy.local'],
    # package_dir={'': 'Phosphorpy'},
    include_package_data=True,
    # package_data={
    #       'config': ['Phosphorpy/local/survey.conf']
    #   },
    url='https://github.com/patrickRauer/Phosphorpy',
    license='GPL',
    author='Patrick Rauer',
    author_email='j.p.rauer@sron.nl',
    description='',
    zip_safe=False
)



from collections import OrderedDict
from astropy import units as u
from threading import Thread
import configparser
from Phosphorpy.external.vizier import Vizier


class MultiDict(OrderedDict):
    _unique = 0   # class variable

    def __setitem__(self, key, val):
        if isinstance(val, dict):
            self._unique += 1
            key += str(self._unique)
        OrderedDict.__setitem__(self, key, val)


class SearchConfig(Thread):
    data = None
    commands = []

    def __init__(self, path=None):
        """
        Class to represent the configuration settings of the search

        :param path: Path to the config-file. Default is None, which means default settings are used.
        :type path: str
        """
        Thread.__init__(self)
        self.commands = configparser.ConfigParser(dict_type=MultiDict, strict=False)
        self.commands.read(path)
        self.commands.sections()

    def __do_query__(self, query_data):
        mode = query_data['mode'].lower()
        if 'circle' in mode or 'box' in mode:
            viz = Vizier(name=query_data['survey'].upper())

            # convert central coordinates
            ra, dec = query_data['center'].split(', ')
            ra = float(ra)
            dec = float(dec)

            # circle mode
            if 'circle' in mode:
                # set the vizier radius with the chosen radius
                viz.radius = float(query_data['radius'])*__get_units__(query_data)
                print(viz.radius)
            # box mode
            elif 'box' in mode:
                r = query_data['radius'].split(', ')
                unit = __get_units__(query_data)
                viz.radius = [r[0]*unit, r[1]*unit]
            else:
                raise AttributeError('{} is not allowed as query mode type.\n' +
                                     'Allowed mode types are \'circle\' or \'box\'.')

            # query around coordinates
            self.data = viz.query_coordinate([ra, dec])
        print('query mode', mode)

    def __create_report__(self, report_data):
        r = Report()
        pass

    def __do_command__(self, command,
                       command_data):
        com = command.lower()
        if 'query' in com:
            print(command_data)
            self.__do_query__(command_data)
        elif 'match' in com:
            pass
        elif 'image' in com:
            pass
        elif 'load' in com:
            pass
        elif 'write' in com or 'save' in com:
            pass
        elif 'report' in com:
            self.__create_report__(command_data)

    def run(self):
        for command in self.commands.sections():
            self.__do_command__(command, dict(self.commands.items(command)))


def __get_units__(query_data):
    """
    Returns the chosen radius unit.
    If no unit is given, it will return degree.

    :param query_data: The query configuration data
    :type query_data: dict
    :return: The chosen unit
    :rtype:
    """
    if 'radius_unit' in query_data.keys():
        unit_string = query_data['radius_unit']
        if 'arcsec' in unit_string:
            return u.arcsec
        elif 'arcmin' in unit_string:
            return u.arcmin
        else:
            return u.deg
    else:
        return u.deg



import numpy as np
import pylab as pl
from astropy.table import Table
from astroquery.vizier import Vizier


class SdssSpec:
    sdss = Vizier(columns=['RA_ICRS', 'DE_ICRS',
                           'umag', 'gmag', 'rmag', 'imag', 'zmag',
                           'subCl'],
                  column_filters={"SpObjID": ">0"})
    sources = None
    stats = None
    m_stats = None

    def __init__(self, row_limit=100000, path=None):
        """
        Class to download and handle SDSS spectroscopic identificated sources

        :param row_limit: Number sources to download
        :type row_limit: int
        :param path: Path to a file with the data from SDSS DR12. Default is None.
        :type path: str
        """
        self.sdss.ROW_LIMIT = row_limit

        if path is not None:
            self.sources = Table.read(path)

    def download_classification(self, cache=True, r_min=20):
        """
        Downloads the classifications and magnitudes of sources from the SDSS DR12 catalog

        :param cache:
            If a path was set, cache=True means that it will use the data from the file.
            False means that it will download a new set. Default is True.
        :type cache: bool
        :param r_min: Minimal brightness in the r-band
        :type r_min: float
        :return: Table with the downloaded data
        :rtype: astropy.table.Table
        """
        try:
            if cache and self.sources is not None:
                return self.sources
            # query the SDSS DR12 for objects with a spectral identification
            # and a set sub-class identification with the
            # astroquery vizier class
            self.sources = self.sdss.query_constraints(catalog='V/147/sdss12',
                                                       SpObjID='>0',
                                                       subCl='!= ',
                                                       rmag='<{}'.format(r_min))[0]
            self.sources = __sub_class_rename__(self.sources)
            self.__stats__()
            return self.sources
        except IndexError:
            return None

    def __stats__(self):
        """
        Counts how often a spectral type appears in the data set

        :return: The counts of the spectral type
        :type: astropy.table.Table
        """
        sub_classes, sub_class_counts = np.unique(self.sources['subCl'], return_counts=True)
        self.stats = Table()
        self.stats['subCl'] = sub_classes
        self.stats['count'] = sub_class_counts
        self.__main_identification_stats__()
        return self.stats

    def __main_identification_stats__(self):
        """
        Returns the main identification of the sources and how often they are in the sample.
        A main identification is for a star the main spectral type.
        Galaxies and QSO's are compressed to only one type (AGN and STARFORMING).
        :return: Dict with the main identifications as keys and the values are how often they are in the sample.
        :rtype: dict
        """
        m_stats = {}
        for s in self.stats:
            sub_cl = s['subCl']

            # if the length is two, it is likely a star identification
            if len(sub_cl) == 2:
                if sub_cl[-1].isdigit():
                    # take only the spectral identification key
                    sub_cl = sub_cl[0]
                    m_stats = __add_counts__(m_stats, sub_cl, s)
                # if the second key is not a number
                else:
                    if 'OB' in sub_cl:
                        sub_cl = 'O'
                    m_stats = __add_counts__(m_stats, sub_cl, s)
            # if the length is larger than two
            else:
                # Compress galaxy and QSO's types
                if sub_cl == 'BROADLINE':
                    sub_cl = "AGN"
                elif sub_cl == 'STARBURST':
                    sub_cl = 'GALAXY'
                elif sub_cl == 'STARFORMING':
                    sub_cl = 'GALAXY'
                elif sub_cl == 'Ldwarf':
                    sub_cl = 'L'
                elif sub_cl == 'Carbon_lines':
                    sub_cl = 'Carbon'
                m_stats = __add_counts__(m_stats, sub_cl, s)
        self.m_stats = m_stats
        return m_stats

    def show_sub_classes(self):
        """
        Plots the distribution of different main spectral identifications
        as a bar plot

        :return:
        """
        pl.clf()
        sp = pl.subplot()
        x = np.zeros((len(self.m_stats), 2))
        for i, k in enumerate(self.m_stats):
            x[i, 0] = i
            x[i, 1] = self.m_stats[k]
        sp.bar(x[:, 0], x[:, 1])
        pl.xticks(x[:, 0], list(self.m_stats.keys()),
                  rotation='vertical')
        sp.set_ylabel('counts')
        sp.set_xlabel('main subclass type')
        sp.set_yscale('log')
        pl.show()


def __add_counts__(m_stats, name, s):
    if name in m_stats.keys():
        m_stats[name] += s['count']
    else:
        m_stats[name] = s['count']
    return m_stats


def __sub_class_rename__(tab):
    """
    Changes the sub-class identification of SDSS DR12 in such a way, that for stars only the main identification
    (Kx, Fy, ...) is left. WD's are also only one sub-class and it won't be distingushed between
    hot and cool ones.
    The identification of galaxies and QSO's is untouched.

    :param tab: The input table with the data from the SDSS DR12
    :type tab: numpy.ndarray, astropy.table.Table
    :return: The input data with the changed sub-class identification
    :rtype: numpy.ndarray, astropy.table.Table
    """
    for i, t in enumerate(tab):
        sub_cl = t['subCl'].split(' ')

        # if there are additional identification numbers
        if len(sub_cl) > 1:
            tab['subCl'][i] = sub_cl[0]

        sub_cl = sub_cl[0]

        # if there is sub spectral identifications. For example K5V change it to K5
        if (len(sub_cl) == 3 or len(sub_cl) == 4) and sub_cl[-1] != 'N':
            if 'sd' in sub_cl:
                sub_cl = sub_cl[2:]
            else:
                sub_cl = sub_cl[:2]
            tab['subCl'][i] = sub_cl
        # if the sub-class is not clear, use just the first one
        elif '/' in sub_cl:
            if 'Ib' in sub_cl:
                sub_cl = sub_cl.split('Ib')[0]
            else:
                sub_cl = sub_cl.split('/')[0]
            tab['subCl'][i] = sub_cl
        # if WD is part of the identification
        elif 'WD' in sub_cl:
            tab['subCl'][i] = 'WD'
        elif 'III' in sub_cl:
            sub_cl = sub_cl[:2]
            tab['subCl'][i] = sub_cl
        elif ':' in sub_cl:
            sub_cl = sub_cl.split(':')[-1]
            tab['subCl'][i] = sub_cl
        elif '.' in sub_cl:
            if 'Ib' in sub_cl:
                sub_cl = sub_cl.split('Ib')[0]
            elif 'V.' in sub_cl:
                sub_cl = sub_cl.split('V.')[0]
            sub_cl = sub_cl.split('.')[0]
            tab['subCl'][i] = sub_cl
        elif 'lb' in sub_cl:
            sub_cl = sub_cl.split('lb')[0]
            tab['subCl'][i] = sub_cl
        elif 'Ib' in sub_cl:
            sub_cl = sub_cl.split('Ib')[0]
            tab['subCl'][i] = sub_cl
        elif 'IV' in sub_cl:
            sub_cl = sub_cl.split('IV')[0]
            tab['subCl'][i] = sub_cl
        elif 'Ia' in sub_cl:
            sub_cl = sub_cl.split('Ia')[0]
            tab['subCl'][i] = sub_cl
        elif 'Vn' in sub_cl:
            sub_cl = sub_cl.split('Vn')[0]
            tab['subCl'][i] = sub_cl

    tab = tab[tab['subCl'] != '']
    return tab
